/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef HSSR_SEQ_COMPRESS_HPP
#define HSSR_SEQ_COMPRESS_HPP

template<typename T, typename S> void
HSS_tree<T,S>::HSSR_seq_Compress(T *A, int *descA, S tol, int dmin, int dmax, int dlim, int dinc, int *deff,
				 int *r,int doRoot, T* Rr, T* Rc, T* Sr, T* Sc, StrumpackRand *rgen,
				 void (*elements)(void *, int *, int *, T *, int *),
				 void (*Amult)(void *, T *, T *, int *, T *, T *, int *),
				 int offset, bool dosync) {
  /* HSS Compression of a dense matrix with randomized sampling.
   * A is a dense n x n matrix.
   * dmin is the number of random vectors to be used; it is used
   * at the beginning of the compression. If, during compression,
   * the HSS rank is found to be too close to the number of
   * random vectors, the number of random vectors is increased.
   * dlim is how close the rank is allowed to be from the number
   * of random vectors.
   * dmax is the maximum number of random vectors to be used.
   * dinc is the increment (number of additional random vectors
   * when expanded).
   * deff is the effective number of random vectors used.
   * r is the maximum rank found during the compression (output).
   * doRoot sets the special treatment of the root node (see below).
   * Rr, Rc, Sr, Sr are random vectors and samplings of A (see below);
   * they are either provided (non-NULL) or computed by the routine.
   * The matrix A can be provided as:
   *  - An explicit matrix,
   *  - An element access routine and explicit samples Sr and Sc,
   *  - An element access routine and a matrix-vector product.
   * When this sequential compression routine is called within
   * a parallel compression, the element access routine needs
   * to be called with global indices (that correspond to the
   * global compression), hence the "offset" parameter. Also,
   * if the access routine is global, we need synchronizations;
   * cf the "dosync" mechanism in the parallel routine. The main
   * difficulty here is that the traversal of the tree has to
   * be level-by-level instead of a postorder. Before and after every
   * level, the process participates to global calls to "elements".
   *
   * Amult is a matrix-matrix multiplication routine that may be
   * provided instead of an explicit matrix.
   *
   * Algorithm (ignoring the random vector expansion mechanism):
   * Step1: Rr and Rc are two n x d random matrices.
   * Step2: Sr and Sc are two n x d samples (resp. row and column) of A.
   * Step3: postorder traversal of the HSS tree; at each node:
   *   a/ Extract a local sample.
   *   b/ Compress with interpolative decomposition.
   *   c/ Update random vectors and indices.
   *
   * The way the root is processed might differ:
   *   - If the tree is part of a larger tree, then the root node should
   *     be processed as any other node (i.e., do a/ b/ c/).
   *   - Otherwise, there is nothing to do (because the local samples
   *     would be zero).
   * Parameter "doRoot" decides whether the root node should be processed
   * as any other node (!=0) or not (0).
   * This is useful when calling this routine as part of a parallel
   * compression for example. doRoot also influences the adaptive
   * sampling mechanism; when doRoot!=0, if there is a failure, the
   * compression simply exits.
   *
   * Limitation: for now the algorithm assumes that the HSS tree is binary.
   */

  int n;
  int d, dupd;
  int i, j, k;
  int ii, kk;
  int ib, ie, br, bc;
  int clean;
  int desc[BLACSCTXTSIZE];
  int ctxt;
  int myid, np;
  int curlevel, nblevels;
  int *level_first, *level_last, *traversal;
  int *tmpI, *tmpJ;
  T *tmp;
  HSS_node<T,S> *child1, *child2;
  double t;
  MPI_Request request;

  int descRS[BLACSCTXTSIZE];

  double tAmult, tBarrier, tElements, tPtc, tRecv; // New

  /* This keeps track of the current state of the node so we know
   * which commands to execute if the node is properly compressed.
   * This information is used in the main for loop for node compression.
   * We remember that UNTOUCHED == 0, PARTIALLY_COMPRESSED == 1, and
   * COMPRESSED == 2.
   */
  int cur_node_state = UNTOUCHED;

#if defined(levelbylevelseq)
  /* Sequential level by level traversal */
  std::cout << "level by level traversal\n" << std::endl;
  dosync=true;
  comm=MPI_COMM_WORLD;
  for(i=0;i<dim_tree;i++) {
    nodes[i]->nb_calls_prev=0;
    nodes[i]->nb_calls_next=0;
  }
#endif

  /* Some initializations */
  n=a_dim;
  *r=0;
  d=dmin;
  time_Rand=0;
  time_GEMM=0;
  time_ID=0;
  time_Amult=0;
  time_Barrier=0;
  time_Elements=0;
  time_Ptc=0;
  time_Recv=0;
  time_samples_GEMM=0;
  time_MPI_Allgather=0;

  np=1;
  if(dosync) {
    MPI_Comm_rank(comm,&myid);
    MPI_Comm_size(comm,&np);
  }
  if(!doRoot) {
    if(np!=1)
      std::cout<<"np!=1 && !doRoot should not happen!" << std::endl;
    comm=MPI_COMM_WORLD;
    for(i=0;i<dim_tree;i++) {
      nodes[i]->nb_calls_prev=0;
      nodes[i]->nb_calls_next=0;
    }
  }

  /* Interaction with the element access routine (parallel interface) */
  if(elements!=NULL) {
    if(!doRoot) {
      blacs_get_(&IZERO,&IZERO,&ctxt);
      blacs_gridinit_(&ctxt,"R",&IONE,&IONE);
    } else {
      ctxt=nodes[dim_tree-1]->context;
    }
  }
  if(dosync) {
    /* Compute level for each node */
    nblevels=0;
    nodes[dim_tree-1]->level=0;
    for(i=dim_tree-1;i>=0;i--)
      if(nodes[i]->nr_children!=0) {
        nodes[i]->children[0]->level=nodes[i]->level+1;
        nodes[i]->children[1]->level=nodes[i]->level+1;
        nblevels=std::max(nblevels,nodes[i]->children[0]->level);
      }
    nblevels++; /* Levels are 0-based */
    /* For each level, compute the first ("leftmost") and
     * the last ("rightmost).
     */
    level_first=new int[nblevels];
    level_last=new int[nblevels];
    for(j=0;j<nblevels;j++) {
      level_first[j]=dim_tree; /* Larger than last index */
      level_last[j]=-1; /* Smaller than first index */
    }
    for(i=0;i<dim_tree;i++) {
      level_first[nodes[i]->level]=std::min(level_first[nodes[i]->level],i);
      level_last[nodes[i]->level]=std::max(level_last[nodes[i]->level],i);
    }
    /* Level by level traversal (bottom-up, and following the PO induced
     * ordering within levels).
     */
    traversal=new int[dim_tree];
    k=0;
    for(j=nblevels-1;j>=0;j--)
      for(i=0;i<dim_tree;i++)
        if(nodes[i]->level==j)
          traversal[k++]=i;
#if defined(debug)
    printMatrix("level_first:"+SSTR(myid),level_first,1,nblevels,1);
    printMatrix("level_last:"+SSTR(myid),level_last,1,nblevels,1);
    printMatrix("traversal:"+SSTR(myid),traversal,1,dim_tree,1);
#endif
  }

  if(Rr==NULL && Rc==NULL && Sr==NULL && Sc==NULL) {
    clean=1;
    /* Step 1: generating random vectors Rr and Rc */
    Rr=new T[n*d];
    Rc=new T[n*d];

    /* Step 2: generating samples: Sr=A*Rr, Sc=A'*Rc */
    /* Check to see if matrix multiplication routine provided */
    Sr=new T[n*d];
    Sc=new T[n*d];
    if(Amult == NULL) {
      rgen= new StrumpackRand(time(NULL));
      t=MPI_Wtime();
      denseRand(Rr,n,d,rgen);
      denseRand(Rc,n,d,rgen);
      time_Rand+=MPI_Wtime()-t;
#if defined(debug)
    printMatrix("Rr",Rr,n,d,n);
    printMatrix("Rc",Rc,n,d,n);
#endif
      t=MPI_Wtime();
      gemm('N','N',n,d,n,ONE,A,n,Rr,n,ZERO,Sr,n);
      gemm('C','N',n,d,n,ONE,A,n,Rc,n,ZERO,Sc,n);
      time_GEMM+=MPI_Wtime()-t;
    } else {
      /* Generate samples using matrix multiplication routine */
      descset_(descRS,&n,&d,&n,&d,&IZERO,&IZERO,&ctxt,&n);
      t=MPI_Wtime();
      Amult(obj,Rr,Rc,descRS,Sr,Sc,descRS);
      time_Amult+=MPI_Wtime()-t;
    }

#if defined(debug)
    printMatrix("Sr",Sr,n,d,n);
    printMatrix("Sc",Sc,n,d,n);
#endif
  } else if (Rr!=NULL && Rc!=NULL && Sr!=NULL && Sc!=NULL) {
    /* Nothing to do, random vectors and samples are provided */
    clean=0;
#if defined(debug)
    printMatrix("Sr",Sr,n,d,n);
    printMatrix("Sc",Sc,n,d,n);
#endif
  } else {
    std::cout << "Error; Rr,Rc,Sr,Sc must be either all NULL or all non-NULL." << std::endl;
    exit(1);
  }

  if(!doRoot)
    /* Start fresh */
    for(i=0;i<dim_tree;i++)
      nodes[i]->state=UNTOUCHED;

  /* Step 3: postorder traversal */
  curlevel=nblevels-1;
  for(ii=0;ii<dim_tree;ii++) {
    if(dosync)
      i=traversal[ii];
    else
      i=ii;
#if defined(debug)
    std::cout << "Node " << i << "; state = " << nodes[i]->state << std::endl;
#endif

    if((nodes[i]->nr_children!=0) && (nodes[i]->nr_children!=2)) {
        std::cout << "Nodes can only have 0 or 2 children. Exiting.\n";
        return;
    }

    /* Save current node state. If not COMPRESSED, then this will be
     * updated during this loop */
    cur_node_state = nodes[i]->state;

#if defined(DOSYNC_ALLGATHER_IJ)
    if(dosync && level_first[curlevel]==i) {

      /*
       * loop over all nodes in this level, in this seq subtree
       *    | sum the sizes of all I and J vectors
       *
       * MPI_Allgather: send your total I,J size to everyone, and get everyone elses
       *
       * Allocate buffer big enough to hold all I,J for all nodes in this level
       * (including those from other processors)
       *
       * loop over all nodes in this level, in this seq subtree
       *    | put I,J in the correct position in the buffer
       *
       * MPI_Allgatherv: send your local part of the big I,J buffer, receive other parts
       *
       * start pointer at first I,J in big IJ buffer (should belong to proc 0)
       * For j=1..nb_call_prev
       *    | get I,J from IJ buffer, advance pointer in IJ buffer
       *    | call elements(I,J,NULL)
       *
       * loop over all nodes in this level, in this seq subtree
       *    | if node is a leaf
       *    |    get I,J from IJ buffer, advance pointer in IJ buffer
       *    |    call elements(I,J,D)
       *    | else
       *    |    get I,J from IJ buffer, advance pointer in IJ buffer
       *    |    call elements(I,J,B12)
       *    |    get I,J from IJ buffer, advance pointer in IJ buffer
       *    |    call elements(I,J,B21)
       *
       * For j=1..nb_call_next
       *    | get I,J from IJ buffer, advance pointer in IJ buffer
       *    | call elements(I,J,NULL)
       *
       */

      int IJsize = 0;
      for(kk=ii; true; kk++) {
        k=traversal[kk];
        if(nodes[k]->state==UNTOUCHED) {
          if(nodes[k]->nr_children==0) { // leaf
            ib=nodes[k]->tree_ib-1;
            ie=nodes[k]->tree_ie-1;
            br=ie-ib+1;
            IJsize += BLACSCTXTSIZE + 2*br; // I,J for D
          } else { // non-leaf
            child1=nodes[k]->children[0];
            child2=nodes[k]->children[1];
            if(child1->state==COMPRESSED && child2->state==COMPRESSED)
              IJsize += 2*BLACSCTXTSIZE + child1->u_rank + child2->v_rank  // I,J for B12
          + child1->v_rank + child2->u_rank;  // I,J for B21
            else IJsize += 2*BLACSCTXTSIZE; // children are not compressed, (B12 & B21 empty)
          }
        }
        if(level_last[curlevel]==k) break;
      }
      int* IJsizes = new int[np];
      // communicate the sizes of all the IJ for each processor
      t=MPI_Wtime();
      MPI_Allgather(&IJsize, 1, MPI_INTEGER, IJsizes, 1, MPI_INTEGER, comm);
      time_MPI_Allgather+=MPI_Wtime()-t;
      int* IJdispls = new int[np];
      IJdispls[0]=0;
      for(j=1;j<np;j++) IJdispls[j] = IJdispls[j-1] + IJsizes[j-1];
      int totIJsizes = IJdispls[np-1] + IJsizes[np-1];
      int* IJ = new int[totIJsizes];
      int* IJpos = IJ + IJdispls[myid];

      for(kk=ii; true; kk++) {
        k=traversal[kk];
        if(nodes[k]->state==UNTOUCHED) {
          if(nodes[k]->nr_children==0) {  // leaf
            ib=nodes[k]->tree_ib-1;
            ie=nodes[k]->tree_ie-1;
            br=ie-ib+1;
            descset_(IJpos,&br,&br,&br,&br,&IZERO,&IZERO,&ctxt,&br);
            IJpos += BLACSCTXTSIZE;
            for(j=0;j<br;j++) IJpos[j]=nodes[k]->tree_ib+j+offset;
            IJpos += br;
            for(j=0;j<br;j++) IJpos[j]=nodes[k]->tree_ib+j+offset;
            IJpos += br;
          } else { // non-leaf
            child1=nodes[k]->children[0];
            child2=nodes[k]->children[1];
            if(child1->state==COMPRESSED && child2->state==COMPRESSED) {
              // B12
              descset_(IJpos,&child1->u_rank,&child2->v_rank,&IONE,&IONE,&IZERO,&IZERO,&ctxt,&child1->u_rank);
              IJpos += BLACSCTXTSIZE;
              for(j=0;j<child1->u_rank;j++) IJpos[j]=child1->Ir[j]+offset;
              IJpos += child1->u_rank;
              for(j=0;j<child2->v_rank;j++) IJpos[j]=child2->Ic[j]+offset;
              IJpos += child2->v_rank;
              // B21
              descset_(IJpos,&child2->u_rank,&child1->v_rank,&IONE,&IONE,&IZERO,&IZERO,&ctxt,&child2->u_rank);
              IJpos += BLACSCTXTSIZE;
              for(j=0;j<child2->u_rank;j++) IJpos[j]=child2->Ir[j]+offset;
              IJpos += child2->u_rank;
              for(j=0;j<child1->v_rank;j++) IJpos[j]=child1->Ic[j]+offset;
              IJpos += child1->v_rank;
            } else { // children are not compressed
              descset_(IJpos,&IZERO,&IZERO,&IONE,&IONE,&IZERO,&IZERO,&ctxt,&IZERO);
              IJpos += BLACSCTXTSIZE;
              descset_(IJpos,&IZERO,&IZERO,&IONE,&IONE,&IZERO,&IZERO,&ctxt,&IZERO);
              IJpos += BLACSCTXTSIZE;
            }
          }
        }
        if(level_last[curlevel]==k) break;
      }
      // communicate all IJ info for this level
      t=MPI_Wtime();
      MPI_Allgatherv(MPI_IN_PLACE, IJsizes[myid], MPI_INTEGER, IJ, IJsizes, IJdispls, MPI_INTEGER, comm);
      time_MPI_Allgather+=MPI_Wtime()-t;
      delete[] IJsizes;
      delete[] IJdispls;

      IJpos = IJ;
      for(j=0; j<nodes[level_first[curlevel]]->nb_calls_prev; j++) {
        if (IJpos[BLACSm] && IJpos[BLACSn]) {
          IJpos[BLACSctxt] = -1;
          tmpI = IJpos + BLACSCTXTSIZE;
          tmpJ = tmpI + IJpos[BLACSm];
          t=MPI_Wtime();
          elements(obj,tmpI,tmpJ,NULL,IJpos);
          time_Elements+=MPI_Wtime()-t;
          ELEMENTS_BARRIER;
        }
        IJpos += BLACSCTXTSIZE + IJpos[BLACSm] + IJpos[BLACSn];
      }

      for(kk=ii; true; kk++) {
        k=traversal[kk];
        if(nodes[k]->state==UNTOUCHED) {
          if(nodes[k]->nr_children==0) {  // leaf
                  ib=nodes[k]->tree_ib-1;
                  ie=nodes[k]->tree_ie-1;
            br=ie-ib+1;
            nodes[k]->D=new T[br*br];
            if(IJpos[BLACSm] && IJpos[BLACSn]) {
              tmpI = IJpos + BLACSCTXTSIZE;
              tmpJ = tmpI + IJpos[BLACSm];
              t=MPI_Wtime();
              elements(obj,tmpI,tmpJ,nodes[k]->D,IJpos);
              time_Elements+=MPI_Wtime()-t;
              ELEMENTS_BARRIER;
            }
            IJpos += BLACSCTXTSIZE + IJpos[BLACSm] + IJpos[BLACSn];
          } else { //non-leaf
            child1=nodes[k]->children[0];
            child2=nodes[k]->children[1];
            if(child1->state==COMPRESSED && child2->state==COMPRESSED) {
              nodes[k]->B12=new T[child1->u_rank*child2->v_rank];
              if(IJpos[BLACSm] && IJpos[BLACSn]) {
                tmpI = IJpos + BLACSCTXTSIZE;
                tmpJ = tmpI + IJpos[BLACSm];
                t=MPI_Wtime();
                elements(obj,tmpI,tmpJ,nodes[k]->B12,IJpos);
                time_Elements+=MPI_Wtime()-t;
                ELEMENTS_BARRIER;
              }
              IJpos += BLACSCTXTSIZE + IJpos[BLACSm] + IJpos[BLACSn];
              nodes[k]->B21=new T[child2->u_rank*child1->v_rank];
              if(IJpos[BLACSm] && IJpos[BLACSn]) {
                tmpI = IJpos + BLACSCTXTSIZE;
                tmpJ = tmpI + IJpos[BLACSm];
                t=MPI_Wtime();
                elements(obj,tmpI,tmpJ,nodes[k]->B21,IJpos);
                time_Elements+=MPI_Wtime()-t;
                ELEMENTS_BARRIER;
              }
              IJpos += BLACSCTXTSIZE + IJpos[BLACSm] + IJpos[BLACSn];
            } else // children are not compressed
              IJpos += 2*BLACSCTXTSIZE;
          }
        }
        if(level_last[curlevel]==k) break;
      }

      for(j=0; j<nodes[level_last[curlevel]]->nb_calls_next; j++) {
        if (IJpos[BLACSm] && IJpos[BLACSn]) {
          IJpos[BLACSctxt] = -1;
          tmpI = IJpos + BLACSCTXTSIZE;
          tmpJ = tmpI + IJpos[BLACSm];
          t=MPI_Wtime();
          elements(obj,tmpI,tmpJ,NULL,IJpos);
          time_Elements+=MPI_Wtime()-t;
          ELEMENTS_BARRIER;
        }
        IJpos += BLACSCTXTSIZE + IJpos[BLACSm] + IJpos[BLACSn];
      }
      delete[] IJ;
    }
#else // NO DOSYNC_ALLGATHER_IJ
    /* If we are in a synchronized traversal, we first perform
     * element extraction as needed
     */
    if(dosync) {
      if(level_first[curlevel]==i) {
        t=MPI_Wtime();
        participate_to_calls(elements,nodes[level_first[curlevel]]->nb_calls_prev,comm,&tBarrier,&tElements,&tRecv);
        tPtc=MPI_Wtime()-t;
        tPtc-=tBarrier;
        tPtc-=tElements;
        tPtc-=tRecv;
        time_Barrier+=tBarrier;
        time_Elements+=tElements;
        time_Ptc+=tPtc;
        time_Recv+=tRecv;

        /* We need to loop through all of the children in this level */
        for(kk=ii; true; kk++) {
          k=traversal[kk];

          if((nodes[k]->nr_children!=0) && (nodes[k]->nr_children!=2)) {
              std::cout << "Nodes can only have 0 or 2 children. Exiting.\n";
              return;
          }

          if((nodes[k]->nr_children==0) ||
             (nodes[k]->nr_children==2 && nodes[k]->children[0]->state==COMPRESSED &&
              nodes[k]->children[1]->state==COMPRESSED)) {

            ib=nodes[k]->tree_ib-1;
            ie=nodes[k]->tree_ie-1;
            /* a/ Extract a local sample with synchronization */
            if(nodes[k]->state==UNTOUCHED)
              dupd=d;
            else
              dupd=dinc;
            if(nodes[k]->nr_children==0) {
              /* Leaf: Sr_loc = Sr(ib:ie,:)-A(ib:ie,ib:ie) *Rr(ib:ie,:)
               *       Sc_loc = Sc(ib:ie,:)-A(ib:ie,ib:ie)'*Rc(ib:ie,:)
               *       with ib,ie the bounds of the clusters at leaf level.
               * Sr_loc and Sc_loc overwrite Sr(ib:ie,:) and Sc(ib:ie,:) */
              br=ie-ib+1;
              bc=br;
              if(nodes[k]->state==UNTOUCHED) {
                /* User-given element access */
                nodes[k]->D=new T[br*br];
                tmpI=new int[br];
                for(j=0;j<br;j++)
                  tmpI[j]=nodes[k]->tree_ib+j+offset;
                tmpJ=tmpI;
                descset_(desc,&br,&br,&br,&br,&IZERO,&IZERO,&ctxt,&br);
                for(j=0;j<np;j++) {
                  if(j!=myid) {
#if defined(debug)
                    std::cout << "dosync: " << myid << " sending information to " << j << std::endl;
#endif
                    MPI_Isend((void*)desc,BLACSCTXTSIZE,MPI_INTEGER,j,2,comm,&request);
                    MPI_Isend((void*)tmpI,br,MPI_INTEGER,j,2,comm,&request);
                    MPI_Isend((void*)tmpJ,br,MPI_INTEGER,j,2,comm,&request);
                  }
                }
                t=MPI_Wtime();
                elements(obj,tmpI,tmpJ,nodes[k]->D,desc);
                time_Elements+=MPI_Wtime()-t;
                t=MPI_Wtime();
                MPI_Barrier(comm);
                time_Barrier+=MPI_Wtime()-t;
                delete[] tmpI;
              }
            } else {
              /* Non-leaf: Sr_loc= [Sr_1-B12 *Rr_2;Sr_2-B21 *Rr_1]
               *           Sc_loc= [Sc_1-B21'*Rc_2;Sc_2-B12'*Rc_1]
               *           with B12=A(Ir_1,Ic_2) and B21=A(Ir_2,Ic_1)
               *           with _1 and _2 the children of i.
               * Sr_loc and Sc_loc overwrite Sr(ib:ib+br-1,:) and Sc(ib:ib+br-1,:) */
              child1=nodes[k]->children[0];
              child2=nodes[k]->children[1];
              br=child1->u_rank+child2->u_rank;
              bc=child1->v_rank+child2->v_rank;
              if(ib!=child1->tree_ib-1 || ie!=child2->tree_ie-1 || child1->tree_ie+1!=child2->tree_ib) {
                std::cout << "Problem with clustering indices." << std::endl;
                exit(1);
              }

              if(nodes[k]->state==UNTOUCHED) {
                /* User-given element access */
                nodes[k]->B12=new T[child1->u_rank*child2->v_rank];
                tmpI=new int[child1->u_rank];
                tmpJ=new int[child2->v_rank];
                for(j=0;j<child1->u_rank;j++)
                  tmpI[j]=child1->Ir[j]+offset;
                for(j=0;j<child2->v_rank;j++)
                  tmpJ[j]=child2->Ic[j]+offset;
                descset_(desc,&child1->u_rank,&child2->v_rank,&IONE,&IONE,&IZERO,&IZERO,&ctxt,&child1->u_rank);
                for(j=0;j<np;j++) {
                  if(j!=myid) {
#if defined(debug)
                    std::cout << "dosync: " << myid << " sending information to " << j << std::endl;
#endif
                    MPI_Isend((void*)desc,BLACSCTXTSIZE,MPI_INTEGER,j,2,comm,&request);
                    MPI_Isend((void*)tmpI,child1->u_rank,MPI_INTEGER,j,2,comm,&request);
                    MPI_Isend((void*)tmpJ,child2->v_rank,MPI_INTEGER,j,2,comm,&request);
                  }
                }
                t=MPI_Wtime();
                elements(obj,tmpI,tmpJ,nodes[k]->B12,desc);
                time_Elements+=MPI_Wtime()-t;
                t=MPI_Wtime();
                MPI_Barrier(comm);
                time_Barrier+=MPI_Wtime()-t;
                delete[] tmpI;
                delete[] tmpJ;

                tmpI=new int[child2->u_rank];
                tmpJ=new int[child1->v_rank];
                for(j=0;j<child2->u_rank;j++)
                  tmpI[j]=child2->Ir[j]+offset;
                for(j=0;j<child1->v_rank;j++)
                  tmpJ[j]=child1->Ic[j]+offset;
                nodes[k]->B21=new T[child2->u_rank*child1->v_rank];
                descset_(desc,&child2->u_rank,&child1->v_rank,&IONE,&IONE,&IZERO,&IZERO,&ctxt,&child2->u_rank);
                for(j=0;j<np;j++) {
                  if(j!=myid) {
#if defined(debug)
                    std::cout << "dosync: " << myid << " sending information to " << j << std::endl;
#endif
                    MPI_Isend((void*)desc,BLACSCTXTSIZE,MPI_INTEGER,j,2,comm,&request);
                    MPI_Isend((void*)tmpI,child2->u_rank,MPI_INTEGER,j,2,comm,&request);
                    MPI_Isend((void*)tmpJ,child1->v_rank,MPI_INTEGER,j,2,comm,&request);
                  }
                }
                t=MPI_Wtime();
                elements(obj,tmpI,tmpJ,nodes[k]->B21,desc);
                time_Elements+=MPI_Wtime()-t;
                t=MPI_Wtime();
                MPI_Barrier(comm);
                time_Barrier+=MPI_Wtime()-t;
                delete[] tmpI;
                delete[] tmpJ;
              }

            }
          } else {
            /* Here, we take care of element extraction when children are
             * not compressed. In this case, we send empty descriptors and
             * do not extract any elements.
             */

            /* We set an empty descriptor */
            descset_(desc,&IZERO,&IZERO,&IONE,&IONE,&IZERO,&IZERO,&ctxt,&IZERO);
            for(j=0;j<np;j++) {
              if(j!=myid) {
#if defined(debug)
                std::cout << "dosync: " << myid << " sending information to " << j << std::endl;
#endif
                /* We send no information to the other processors */
                MPI_Isend((void*)desc,BLACSCTXTSIZE,MPI_INTEGER,j,2,comm,&request);
                MPI_Isend((void*)NULL,0,MPI_INTEGER,j,2,comm,&request);
                MPI_Isend((void*)NULL,0,MPI_INTEGER,j,2,comm,&request);
              }
            }
            t=MPI_Wtime();
            elements(obj,NULL,NULL,nodes[k]->B12,desc);
            time_Elements+=MPI_Wtime()-t;
            t=MPI_Wtime();
            MPI_Barrier(comm);
            time_Barrier+=MPI_Wtime()-t;

            /* We set an empty descriptor */
            descset_(desc,&IZERO,&IZERO,&IONE,&IONE,&IZERO,&IZERO,&ctxt,&IZERO);
            for(j=0;j<np;j++) {
              if(j!=myid) {
#if defined(debug)
                std::cout << "dosync: " << myid << " sending information to " << j << std::endl;
#endif
                /* We send no information to the other processors */
                MPI_Isend((void*)desc,BLACSCTXTSIZE,MPI_INTEGER,j,2,comm,&request);
                MPI_Isend((void*)NULL,0,MPI_INTEGER,j,2,comm,&request);
                MPI_Isend((void*)NULL,0,MPI_INTEGER,j,2,comm,&request);
              }
            }
            t=MPI_Wtime();
            elements(obj,NULL,NULL,nodes[k]->B21,desc);
            time_Elements+=MPI_Wtime()-t;
            t=MPI_Wtime();
            MPI_Barrier(comm);
            time_Barrier+=MPI_Wtime()-t;
          }

          /* We break out of the loop when we hit the last node in the level */
          if(level_last[curlevel]==k)
            break;
        }

        /* Need to look into this */
        t=MPI_Wtime();
        participate_to_calls(elements,nodes[level_last[curlevel]]->nb_calls_next,comm,&tBarrier,&tElements,&tRecv);
        tPtc=MPI_Wtime()-t;
        tPtc-=tBarrier;
        tPtc-=tElements;
        tPtc-=tRecv;
        time_Barrier+=tBarrier;
        time_Elements+=tElements;
        time_Ptc+=tPtc;
        time_Recv+=tRecv;
      }
    }
#endif // END NO DOSYNC_ALLGATHER_IJ


    /* We only attempt compression if we are at a leaf or if both of the node's
     * children are compress */
    if((nodes[i]->nr_children==0) ||
       (nodes[i]->nr_children==2 && nodes[i]->children[0]->state==COMPRESSED &&
        nodes[i]->children[1]->state==COMPRESSED)) {

      ib=nodes[i]->tree_ib-1;
      ie=nodes[i]->tree_ie-1;
      /* a/ Extract a local sample */
      if(nodes[i]->state==UNTOUCHED)
        dupd=d;
      else
        dupd=dinc;
      if(nodes[i]->nr_children==0) {
        /* Leaf: Sr_loc = Sr(ib:ie,:)-A(ib:ie,ib:ie) *Rr(ib:ie,:)
         *       Sc_loc = Sc(ib:ie,:)-A(ib:ie,ib:ie)'*Rc(ib:ie,:)
         *       with ib,ie the bounds of the clusters at leaf level.
         * Sr_loc and Sc_loc overwrite Sr(ib:ie,:) and Sc(ib:ie,:) */
        br=ie-ib+1;
        bc=br;
        if(nodes[i]->state==UNTOUCHED) {
          if(A!=NULL) {
            /* Explicit matrix */
            extractSubmatrix(&A[ib*n+ib],br,br,n,NULL,-1,NULL,-1,&(nodes[i]->D));
          } else {
            /* User-given element access */
            if(!dosync) {
              /* We only need to worry about element extraction if the
               * element access routine does not require synchronization,
               * because if it does, we have already extracted all of the
               * elements that we need for this level
               */
              nodes[i]->D=new T[br*br];
              tmpI=new int[br];
              for(j=0;j<br;j++)
                tmpI[j]=nodes[i]->tree_ib+j+offset;
              tmpJ=tmpI;
              descset_(desc,&br,&br,&br,&br,&IZERO,&IZERO,&ctxt,&br);
              t=MPI_Wtime();
              elements(obj,tmpI,tmpJ,nodes[i]->D,desc);
              time_Elements+=MPI_Wtime()-t;
              delete[] tmpI;
            }
          }
        }
        if(i==dim_tree-1 && !doRoot) {
          nodes[i]->state=COMPRESSED;
          break;
        }

        t=MPI_Wtime();
        gemm('N','N',br,dupd,br,NONE,nodes[i]->D,br,&Rr[ib+(d-dupd)*n],n,ONE,&Sr[ib+(d-dupd)*n],n);
        gemm('C','N',br,dupd,br,NONE,nodes[i]->D,br,&Rc[ib+(d-dupd)*n],n,ONE,&Sc[ib+(d-dupd)*n],n);
        time_samples_GEMM+=MPI_Wtime()-t;
      } else {
        /* Non-leaf: Sr_loc= [Sr_1-B12 *Rr_2;Sr_2-B21 *Rr_1]
         *           Sc_loc= [Sc_1-B21'*Rc_2;Sc_2-B12'*Rc_1]
         *           with B12=A(Ir_1,Ic_2) and B21=A(Ir_2,Ic_1)
         *           with _1 and _2 the children of i.
         * Sr_loc and Sc_loc overwrite Sr(ib:ib+br-1,:) and Sc(ib:ib+br-1,:) */
        child1=nodes[i]->children[0];
        child2=nodes[i]->children[1];
        br=child1->u_rank+child2->u_rank;
        bc=child1->v_rank+child2->v_rank;
        if(ib!=child1->tree_ib-1 || ie!=child2->tree_ie-1 || child1->tree_ie+1!=child2->tree_ib) {
          std::cout << "Problem with clustering indices." << std::endl;
          exit(1);
        }

        if(nodes[i]->state==UNTOUCHED) {
          if(A!=NULL) {
            /* Explicit matrix */
            extractSubmatrix(A,n,n,n,child1->Ir,child1->u_rank,child2->Ic,child2->v_rank,&nodes[i]->B12);
            extractSubmatrix(A,n,n,n,child2->Ir,child2->u_rank,child1->Ic,child1->v_rank,&nodes[i]->B21);
          } else {
            /* User-given element access */
            if(!dosync) {
              /* We only need to worry about element extraction if the
               * element access routine does not require synchronization,
               * because if it does, we have already extracted all of the
               * elements that we need for this level
               */
              nodes[i]->B12=new T[child1->u_rank*child2->v_rank];
              tmpI=new int[child1->u_rank];
              tmpJ=new int[child2->v_rank];
              for(j=0;j<child1->u_rank;j++)
                tmpI[j]=child1->Ir[j]+offset;
              for(j=0;j<child2->v_rank;j++)
                tmpJ[j]=child2->Ic[j]+offset;
              descset_(desc,&child1->u_rank,&child2->v_rank,&IONE,&IONE,&IZERO,&IZERO,&ctxt,&child1->u_rank);
              t=MPI_Wtime();
              elements(obj,tmpI,tmpJ,nodes[i]->B12,desc);
              time_Elements+=MPI_Wtime()-t;
              delete[] tmpI;
              delete[] tmpJ;

              tmpI=new int[child2->u_rank];
              tmpJ=new int[child1->v_rank];
              for(j=0;j<child2->u_rank;j++)
                tmpI[j]=child2->Ir[j]+offset;
              for(j=0;j<child1->v_rank;j++)
                tmpJ[j]=child1->Ic[j]+offset;
              nodes[i]->B21=new T[child2->u_rank*child1->v_rank];
              descset_(desc,&child2->u_rank,&child1->v_rank,&IONE,&IONE,&IZERO,&IZERO,&ctxt,&child2->u_rank);
              t=MPI_Wtime();
              elements(obj,tmpI,tmpJ,nodes[i]->B21,desc);
              time_Elements+=MPI_Wtime()-t;
              delete[] tmpI;
              delete[] tmpJ;
            }
          }
        }

        if(i==dim_tree-1 && !doRoot) {
          nodes[i]->state=COMPRESSED;
          break;
        }

        if(child1->u_rank && child2->v_rank) {
          t=MPI_Wtime();
          gemm('N','N',child1->u_rank,dupd,child2->v_rank,NONE,nodes[i]->B12,child1->u_rank,&Rr[child2->tree_ib-1+(d-dupd)*n],n,ONE,&Sr[child1->tree_ib-1+(d-dupd)*n],n);
          time_samples_GEMM+=MPI_Wtime()-t;
        }
        if(child2->u_rank && child1->v_rank) {
          t=MPI_Wtime();
          gemm('N','N',child2->u_rank,dupd,child1->v_rank,NONE,nodes[i]->B21,child2->u_rank,&Rr[child1->tree_ib-1+(d-dupd)*n],n,ONE,&Sr[child2->tree_ib-1+(d-dupd)*n],n);
          time_samples_GEMM+=MPI_Wtime()-t;
        }
        /* Move Sr_child2 below Sr_child1 line by line */
        for(k=0;k<child2->u_rank;k++)
          for(j=d-dupd;j<d;j++)
            Sr[n*j+ib+child1->u_rank+k]=Sr[n*j+child2->tree_ib-1+k];

        if(child1->v_rank && child2->u_rank) {
          t=MPI_Wtime();
          gemm('C','N',child1->v_rank,dupd,child2->u_rank,NONE,nodes[i]->B21,child2->u_rank,&Rc[child2->tree_ib-1+(d-dupd)*n],n,ONE,&Sc[child1->tree_ib-1+(d-dupd)*n],n);
          time_samples_GEMM+=MPI_Wtime()-t;
        }
        if(child2->v_rank && child1->u_rank) {
          t=MPI_Wtime();
          gemm('C','N',child2->v_rank,dupd,child1->u_rank,NONE,nodes[i]->B12,child1->u_rank,&Rc[child1->tree_ib-1+(d-dupd)*n],n,ONE,&Sc[child2->tree_ib-1+(d-dupd)*n],n);
          time_samples_GEMM+=MPI_Wtime()-t;
        }
        /* Move Sc_child2 below Sc_child1 line by line */
        for(k=0;k<child2->v_rank;k++)
          for(j=d-dupd;j<d;j++)
            Sc[n*j+ib+child1->v_rank+k]=Sc[n*j+child2->tree_ib-1+k];
      }


      /* b/ Compress with Interpolative Decomposition: define the generators U and V
       * and the compressed indices Ir and Ic (resp. row and column indices).
       * Indices in output of ID are local to the matrix passed to the routine;
       * they are shifted in c/
       *
       * Sr/Sc are transposed into tmp before compression.
       *
       * After compression, Sr/Sc=Sr/Sc(Ir/Ic,:).
       */
      if(nodes[i]->state!=COMPRESSED) {
        /* We attempt compression */
        tmp=new T[d*br];
        copyTrans(br,d,&Sr[ib],n,tmp,d);
        t=MPI_Wtime();
        seq_XInterpol(d,br,tmp,tol,&(nodes[i]->U_E),&(nodes[i]->u_rank),&(nodes[i]->Ir),&(nodes[i]->U_perm));
        time_ID+=MPI_Wtime()-t;
        delete[] tmp;

        if(d-nodes[i]->u_rank>=dlim) {
          /* If the u_rank is small enough, then we compute V and v_rank.
           * This saves us from performing more Interpolative Decompositions
           * than necessary. */
          tmp=new T[d*bc];
          copyTrans(bc,d,&Sc[ib],n,tmp,d);
          t=MPI_Wtime();
          seq_XInterpol(d,bc,tmp,tol,&(nodes[i]->V_E),&(nodes[i]->v_rank),&(nodes[i]->Ic),&(nodes[i]->V_perm));
          time_ID+=MPI_Wtime()-t;
          delete[] tmp;

          /* We now assume that node has been successfully compressed.
           * If this is not the case, the next statement will fix that. */
          cur_node_state=COMPRESSED;
        }

        if(((d-nodes[i]->u_rank<dlim) || (d-nodes[i]->v_rank<dlim)) &&
            d+dinc<=dmax) {
          /* Reject this compression; we will eventually need to
           * add new random vectors and restart the compression, but we
           * continue through the nodes in order to compress as many nodes
           * as possible with the current number of random samples. */
          nodes[i]->state=PARTIALLY_COMPRESSED;
          delete[] nodes[i]->U_E;
          delete[] nodes[i]->Ir;
          delete[] nodes[i]->U_perm;
          nodes[i]->u_rank=-1;
          if(d-nodes[i]->u_rank>=dlim) {
            /* This will only need to be deleted in the (rare) case that
             * the u_rank is small enough but the v_rank is not */
            delete[] nodes[i]->V_E;
            delete[] nodes[i]->Ic;
            delete[] nodes[i]->V_perm;
            nodes[i]->v_rank=-1;
          }

          /* We know that this node failed to compress, so we
           * do not continue in the compression routine sequence */
          cur_node_state=PARTIALLY_COMPRESSED;
        }

        if(cur_node_state==COMPRESSED) {
          /* If the node has been successfully compressed, we proceed
           * as normal. Otherwise, we do not do anything. */
          *r=std::max(nodes[i]->u_rank,*r);
          nodes[i]->u_rows=br-nodes[i]->u_rank;
          tmp=new T[nodes[i]->u_rows*nodes[i]->u_rank];
          copyTrans(nodes[i]->u_rank,nodes[i]->u_rows,nodes[i]->U_E,nodes[i]->u_rank,tmp,nodes[i]->u_rows);
          delete[] nodes[i]->U_E;
          nodes[i]->U_E=tmp;
          extractSubmatrix(&Sr[ib],nodes[i]->u_rank,d,n,nodes[i]->Ir,nodes[i]->u_rank,NULL,-1,&tmp);
          lacpy('N',nodes[i]->u_rank,d,tmp,nodes[i]->u_rank,&Sr[ib],n);
          delete[] tmp;
          nodes[i]->Jr=new int[nodes[i]->u_rank];
          for(j=0;j<nodes[i]->u_rank;j++)
            nodes[i]->Jr[j]=nodes[i]->Ir[j];

          *r=std::max(nodes[i]->v_rank,*r);
          nodes[i]->v_rows=bc-nodes[i]->v_rank;
          tmp=new T[nodes[i]->v_rows*nodes[i]->v_rank];
          copyTrans(nodes[i]->v_rank,nodes[i]->v_rows,nodes[i]->V_E,nodes[i]->v_rank,tmp,nodes[i]->v_rows);
          delete[] nodes[i]->V_E;
          nodes[i]->V_E=tmp;
          extractSubmatrix(&Sc[ib],nodes[i]->v_rank,d,n,nodes[i]->Ic,nodes[i]->v_rank,NULL,-1,&tmp);
          lacpy('N',nodes[i]->v_rank,d,tmp,nodes[i]->v_rank,&Sc[ib],n);
          delete[] tmp;
          nodes[i]->Jc=new int[nodes[i]->v_rank];
          for(j=0;j<nodes[i]->v_rank;j++)
            nodes[i]->Jc[j]=nodes[i]->Ic[j];
        }
      } else {
        /* An ancestor of i restarted the traversal. We update the new columns of Sr/Sc; we
         * have to be careful because Ir/Ic have been previously transformed to absolute indices.
         * We use Jr/Jc instead. */
        extractSubmatrix(&Sr[ib+(d-dinc)*n],nodes[i]->u_rank,dinc,n,nodes[i]->Jr,nodes[i]->u_rank,NULL,-1,&tmp);
        lacpy('N',nodes[i]->u_rank,dinc,tmp,nodes[i]->u_rank,&Sr[ib+(d-dinc)*n],n);
        delete[] tmp;

        extractSubmatrix(&Sc[ib+(d-dinc)*n],nodes[i]->v_rank,dinc,n,nodes[i]->Jc,nodes[i]->v_rank,NULL,-1,&tmp);
        lacpy('N',nodes[i]->v_rank,dinc,tmp,nodes[i]->v_rank,&Sc[ib+(d-dinc)*n],n);
        delete[] tmp;
      }


      /* c/ Update random vectors and shift indices */
      if(cur_node_state==COMPRESSED) {
        /* We proceed as normal. Otherwise, we update the node state */
        if(nodes[i]->state==COMPRESSED)
          dupd=dinc;
        else
          dupd=d;
        if(nodes[i]->nr_children==0) {
          /* Leaf: Rr_tau=V'*Rr(ib:ie,:)
           *       Rc_tau=U'*Rc(ib:ie,:)
           *       Ir_tau=Ir_tau(Jr) (with Jr stored in Ir_tau already)
           *       Ic_tau=Ic_tau(Jc) (with Jc stored in Ic_tau already)
           * Rr_tau and Rc_tau overwrite Rr and Rc */
          nodes[i]->seq_applyGenerator('C','V',dupd,ONE,&Rr[ib+(d-dupd)*n],n,ZERO,&Rr[ib+(d-dupd)*n],n);
          if(nodes[i]->state!=COMPRESSED)
            for(j=0;j<nodes[i]->u_rank;j++)
              nodes[i]->Ir[j]+=ib;
          nodes[i]->seq_applyGenerator('C','U',dupd,ONE,&Rc[ib+(d-dupd)*n],n,ZERO,&Rc[ib+(d-dupd)*n],n);
          if(nodes[i]->state!=COMPRESSED)
            for(j=0;j<nodes[i]->v_rank;j++)
              nodes[i]->Ic[j]+=ib;
        } else {
          /* Non-leaf: Rr_tau=V'*[Rr_1; Rr_2]
           *           Rc_tau=U'*[Rc_1; Rc_2]
           *           Ir_tau=[Ir_1 Ir_2](Jr)
           *           Ic_tau=[Ic_1 Ic_2](Jc)
           */
          tmp=new T[bc*dupd];
          lacpy('N',child1->v_rank,dupd,&Rr[child1->tree_ib-1+(d-dupd)*n],n,tmp,bc);
          lacpy('N',child2->v_rank,dupd,&Rr[child2->tree_ib-1+(d-dupd)*n],n,&tmp[child1->v_rank],bc);
          nodes[i]->seq_applyGenerator('C','V',dupd,ONE,tmp,bc,ZERO,&Rr[ib+(d-dupd)*n],n);
          delete[] tmp;
          if(nodes[i]->state!=COMPRESSED)
            for(j=0;j<nodes[i]->u_rank;j++)
              if(nodes[i]->Ir[j]<=child1->u_rank)
                nodes[i]->Ir[j]=child1->Ir[nodes[i]->Ir[j]-1];
              else
                nodes[i]->Ir[j]=child2->Ir[nodes[i]->Ir[j]-child1->u_rank-1];

          tmp=new T[br*dupd];
          lacpy('N',child1->u_rank,dupd,&Rc[child1->tree_ib-1+(d-dupd)*n],n,tmp,br);
          lacpy('N',child2->u_rank,dupd,&Rc[child2->tree_ib-1+(d-dupd)*n],n,&tmp[child1->u_rank],br);
          nodes[i]->seq_applyGenerator('C','U',dupd,ONE,tmp,br,ZERO,&Rc[ib+(d-dupd)*n],n);
          delete[] tmp;
          if(nodes[i]->state!=COMPRESSED)
            for(j=0;j<nodes[i]->v_rank;j++)
              if(nodes[i]->Ic[j]<=child1->v_rank)
                nodes[i]->Ic[j]=child1->Ic[nodes[i]->Ic[j]-1];
              else
                nodes[i]->Ic[j]=child2->Ic[nodes[i]->Ic[j]-child1->v_rank-1];
        }
        nodes[i]->numrand=d;
        nodes[i]->state=COMPRESSED;
      } else {
        nodes[i]->state=PARTIALLY_COMPRESSED;
      }
    }

    if(dosync)
      if(level_last[curlevel]==i)
        curlevel--;

    if(i==(dim_tree-1)) {
      /* We do not want to restart the tree until we are at the
       * top. This allows us to make as much progress as we can with
       * the fewest samples. */

      /* If tree is a sequential subtree of a parallel execution, then
       * we already know the current state and need to leave the loop.
       * If more vectors are needed, they will be computed at the top
       * of the tree.
       */
      if(doRoot)
        break;

      /* Add more columns to Rr and Sr for a restart of sequential
       * compression */
      /* Add more columns to Rc and Sc for a restart of sequential
       * compression */

      d+=dinc;
      tmp=new T[n*d]();
      lacpy('N',n,d-dinc,Rr,n,tmp,n);
      delete[] Rr;
      Rr=tmp;
      tmp=new T[n*d]();
      lacpy('N',n,d-dinc,Rc,n,tmp,n);
      delete[] Rc;
      Rc=tmp;
      tmp=new T[n*d]();
      lacpy('N',n,d-dinc,Sr,n,tmp,n);
      delete[] Sr;
      Sr=tmp;
      tmp=new T[n*d]();
      lacpy('N',n,d-dinc,Sc,n,tmp,n);
      delete[] Sc;
      Sc=tmp;
      if(Amult == NULL) {
	t=MPI_Wtime();
	denseRand(&Rr[(d-dinc)*n],n,dinc,rgen);
	denseRand(&Rc[(d-dinc)*n],n,dinc,rgen);
	time_Rand+=MPI_Wtime()-t;
	t=MPI_Wtime();
        gemm('N','N',n,dinc,n,ONE,A,n,&Rr[(d-dinc)*n],n,ZERO,&Sr[(d-dinc)*n],n);
        gemm('C','N',n,dinc,n,ONE,A,n,&Rc[(d-dinc)*n],n,ZERO,&Sc[(d-dinc)*n],n);
        time_GEMM+=MPI_Wtime()-t;
      } else {
	// Rr and Rc are set by the Amult user routine
	descset_(descRS,&n,&dinc,&n,&dinc,&IZERO,&IZERO,&ctxt,&n);
        /* We need to update the descRS to account for the new vectors */
        t=MPI_Wtime();
	Amult(obj,Rr+(d-dinc)*n,Rc+(d-dinc)*n,descRS,Sr+(d-dinc)*n,Sc+(d-dinc)*n,descRS);
        time_Amult+=MPI_Wtime()-t;
      }

      /* Restart the sequential compression routine */
      ii=-1;
      continue;
    }

#if defined(debug)
    std::cout<< "u_rank=" << nodes[i]->u_rank <<  "; u_rows=" << nodes[i]->u_rows << std::endl;
    std::cout<< "v_rank=" << nodes[i]->v_rank <<  "; v_rows=" << nodes[i]->v_rows << std::endl;
    if(nodes[i]->nr_children==0)
      printMatrix("D",nodes[i]->D,br,bc,br);
    else {
      printMatrix("B12",nodes[i]->B12,child1->u_rank,child2->v_rank,child1->u_rank);
      printMatrix("B21",nodes[i]->B21,child2->u_rank,child1->v_rank,child2->u_rank);
    }
    printMatrix("U_E",nodes[i]->U_E,nodes[i]->u_rows,nodes[i]->u_rank,nodes[i]->u_rows);
    printMatrix("U_P",nodes[i]->U_perm,1,nodes[i]->u_rank+nodes[i]->u_rows,1);
    printMatrix("Ir",nodes[i]->Ir,1,nodes[i]->u_rank,1);
    printMatrix("V_E",nodes[i]->V_E,nodes[i]->v_rows,nodes[i]->v_rank,nodes[i]->v_rows);
    printMatrix("V_P",nodes[i]->V_perm,1,nodes[i]->v_rank+nodes[i]->v_rows,1);
    printMatrix("Ic",nodes[i]->Ic,1,nodes[i]->v_rank,1);
    printMatrix("Rr",&Rr[ib],nodes[i]->v_rank,d,n);
    printMatrix("Sr",&Sr[ib],nodes[i]->u_rank,d,n);
    printMatrix("Rc",&Rc[ib],nodes[i]->u_rank,d,n);
    printMatrix("Sc",&Sc[ib],nodes[i]->v_rank,d,n);
#endif
  } 

#if defined(debug)
  if(!doRoot) {
    i=dim_tree-1;
    ib=nodes[i]->tree_ib-1;
    ie=nodes[i]->tree_ie-1;
    br=ie-ib+1;
    if(nodes[i]->nr_children==0)
      printMatrix("D",nodes[i]->D,br,br,br);
    else {
      child1=nodes[i]->children[0];
      child2=nodes[i]->children[1];
      printMatrix("B12",nodes[i]->B12,child1->u_rank,child2->v_rank,child1->u_rank);
      printMatrix("B21",nodes[i]->B21,child2->u_rank,child1->v_rank,child2->u_rank);
    }
  }
#endif

  *deff=d;

  if(clean) {
    /* Delete only random vector and samples if they were not provided */
    delete[] Rr;
    delete[] Rc;
    delete[] Sr;
    delete[] Sc;
  }

  if(dosync) {
    delete[] level_first;
    delete[] level_last;
    delete[] traversal;
  }

}

#endif // HSSR_SEQ_COMPRESS_HPP
