/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef HSS_PAR_GENERATESCHUR_HPP
#define HSS_PAR_GENERATESCHUR_HPP

template<typename T, typename S> void HSS_tree<T,S>::HSS_par_GenerateSchur(gridinfo *ginfo) {
  /* Assuming the matrix is already compressed and partially
   * factored, build the generators theta and psi s.t.
   * the Schur complement S is S=A22 - theta Vtilda_k^* psi, with
   * A22 (submatrix of the original matrix A) in compressed
   * form and theta and psi two dense matrices.
   *
   * With k the left child of the root node and w the right
   * child of the root node:
   * theta = U_q^big B_qk
   * psi = Utilda_k^-1 Ltilda_k^-1 Ptilda_k U_k B_kq V_q^big^*
   *
   * The low-rank Schur complement update is represented as
   * theta Vtilda_k^* psi.
   *
   * The parallel version follows the same flow as the parallel compression;
   * there are three kinds of nodes to distinguish: serial subtrees, leaves
   * that are mapped on several processes, and parallel interior nodes.
   * The main difference is that we traverse only the subtree rooted at q.
   *
   */
  int mb, nb;
  int locr, locc;
  int myid;
  int br, bc;
  int qnode;
  int istart, iend;
  int i, j, k;
  int ch1, ch2;
  int idummy, ierr;
  int proc_b, proc_g;
  int proc_bc, proc_gc;
  int *children1, *children2;
  int *rowsUbig, *rowsVbig;
  int descUe[BLACSCTXTSIZE], descVe[BLACSCTXTSIZE];
  int desc_ch1[BLACSCTXTSIZE], desc_ch2[BLACSCTXTSIZE];
  int desc1[BLACSCTXTSIZE], desc2[BLACSCTXTSIZE];
  int desc[BLACSCTXTSIZE];
  HSS_node<T,S> *child1, *child2;
  T *Ue, *Ve, *Xbig1, *Xbig2, *tmp=NULL;
  T **Ubig, **Vbig;

  MPI_Comm_rank(comm,&myid);

  mb=ginfo->mb;
  nb=ginfo->nb;

  /* We compute U_q^big and V_q^big by traversing the subtree rooted at q
   * and multiplying out the generators (we "unfold" the nested basis).
   * Here U_bug and V_big are not fields  of a node but arrays indexed following the
   * postorder; the pointers "children" of a node are not enough to find the
   * right in y and z, what we need is numbers in the postorder. "children1"
   * and "children2" are the same information as children but they provide
   * numbers (indices in the tree).
   *
   * children1: children[i]==j: node number j is left child of node number i.
   * child1: pointer to left child (type HSS_node) of node nodes[i].
   */

  children1=new int[dim_tree];
  for(i=0;i<dim_tree;i++) {
    children1[i]=-1;
    if(nodes[i]->nr_children!=0)
      for(j=0;j<dim_tree;j++)
        if(nodes[j]==nodes[i]->children[0]) {
          children1[i]=j;
          break;
        }
  }
  children2=new int[dim_tree];
  for(i=0;i<dim_tree;i++) {
    children2[i]=-1;
    if(nodes[i]->nr_children!=0)
      for(j=0;j<dim_tree;j++)
        if(nodes[j]==nodes[i]->children[1]) {
          children2[i]=j;
          break;
        }
  }
  Ubig=new T*[dim_tree];
  for(i=0;i<dim_tree;i++)
    Ubig[i]=NULL;
  rowsUbig=new int[dim_tree];
  Vbig=new T*[dim_tree];
  for(i=0;i<dim_tree;i++)
    Vbig[i]=NULL;
  rowsVbig=new int[dim_tree];

  /* The right child of the root node is its predecessor in the postorder */
  qnode=dim_tree-2;

  /* Serial subtree: traverse only if it is below q */
  if(subtree[2*myid]>=0 && subtree[2*myid+1]>knode) {
    istart=subtree[2*myid];
    iend=subtree[2*myid+1];
    for(i=istart;i<=iend;i++) {
      /* Traverse subtree */
      if(nodes[i]->nr_children==0) {
        br=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
        bc=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
      } else {
        child1=nodes[i]->children[0];
        child2=nodes[i]->children[1];
        ch1=children1[i];
        ch2=children2[i];
        br=child1->u_rank+child2->u_rank;
        bc=child1->v_rank+child2->v_rank;
      }
      /* We expand U into its explicit form U=P*[I;E] */
      Ue=new T[br*nodes[i]->u_rank]();
      for(j=0;j<nodes[i]->u_rank;j++)
        Ue[j*br+j]=ONE;
      lacpy('N',nodes[i]->u_rows,nodes[i]->u_rank,nodes[i]->U_E,nodes[i]->u_rows,&Ue[nodes[i]->u_rank],br);
      for(k=br-1;k>=0;k--)
        if(k!=nodes[i]->U_perm[k]-1)
          swap(nodes[i]->u_rank,&Ue[k],br,&Ue[nodes[i]->U_perm[k]-1],br);

      /* We expand V into its explicit form V=P*[I;E] */
      Ve=new T[bc*nodes[i]->v_rank]();
      for(j=0;j<nodes[i]->v_rank;j++)
        Ve[j*bc+j]=ONE;
      lacpy('N',nodes[i]->v_rows,nodes[i]->v_rank,nodes[i]->V_E,nodes[i]->v_rows,&Ve[nodes[i]->v_rank],bc);
      for(k=bc-1;k>=0;k--)
        if(k!=nodes[i]->V_perm[k]-1)
          swap(nodes[i]->v_rank,&Ve[k],bc,&Ve[nodes[i]->V_perm[k]-1],bc);

      if(nodes[i]->nr_children==0) {
        /* Leaf: U^big=U, V^big=V, stored in explicit form */
        Ubig[i]=Ue;
        rowsUbig[i]=br;
        Vbig[i]=Ve;
        rowsVbig[i]=bc;
      } else {
        /* Interior node: U^big = diag(U^big_child1,U^big_child2) * U, ditto for V */
        rowsUbig[i]=rowsUbig[ch1]+rowsUbig[ch2];
        Ubig[i]=new T[rowsUbig[i]*nodes[i]->u_rank];
        if(rowsUbig[ch1] && br && rowsUbig[i])
          gemm('N','N',rowsUbig[ch1],nodes[i]->u_rank,nodes[ch1]->u_rank,ONE,Ubig[ch1],rowsUbig[ch1],Ue,br,ZERO,Ubig[i],rowsUbig[i]);
        if(rowsUbig[ch2] && br && rowsUbig[i])
          gemm('N','N',rowsUbig[ch2],nodes[i]->u_rank,nodes[ch2]->u_rank,ONE,Ubig[ch2],rowsUbig[ch2],&Ue[nodes[ch1]->u_rank],br,ZERO,&Ubig[i][rowsUbig[ch1]],rowsUbig[i]);
        delete[] Ue;
        delete[] Ubig[ch1];
        delete[] Ubig[ch2];

        rowsVbig[i]=rowsVbig[ch1]+rowsVbig[ch2];
        Vbig[i]=new T[rowsVbig[i]*nodes[i]->v_rank]();
        if(rowsVbig[ch1] && bc && rowsVbig[i])
          gemm('N','N',rowsVbig[ch1],nodes[i]->v_rank,nodes[ch1]->v_rank,ONE,Vbig[ch1],rowsVbig[ch1],Ve,bc,ZERO,Vbig[i],rowsVbig[i]);
        if(rowsVbig[ch2] && bc && rowsVbig[i])
          gemm('N','N',rowsVbig[ch2],nodes[i]->v_rank,nodes[ch2]->v_rank,ONE,Vbig[ch2],rowsVbig[ch2],&Ve[nodes[ch1]->v_rank],bc,ZERO,&Vbig[i][rowsVbig[ch1]],rowsVbig[i]);
        delete[] Ve;
        delete[] Vbig[ch1];
        delete[] Vbig[ch2];
      }
    }
  }

  /* Parallel leaves */
  for(i=knode+1;i<=qnode;i++) {
    br=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
    bc=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
    if(   nodes[i]->procs[1]>nodes[i]->procs[0]
       && nodes[i]->nr_children==0
       && nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1>=myid
       && nodes[i]->procs[0]<=myid
      ) {
      /* We expand U into its explicit form U=P*[I;E] */
      locr=numroc_(&br,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
      locc=numroc_(&nodes[i]->u_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
      Ue=new T[locr*locc]();
      idummy=std::max(1,locr);
      descinit_(descUe,&br,&nodes[i]->u_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
      for(k=1;k<=nodes[i]->u_rank;k++)
        if(nodes[i]->myrow==indxg2p_(&k,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow) && nodes[i]->mycol==indxg2p_(&k,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol))
          Ue[(indxg2l_(&k,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol)-1)*locr+indxg2l_(&k,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow)-1]=ONE;
      pgeadd('N',nodes[i]->u_rows,nodes[i]->u_rank,ONE,nodes[i]->U_E,IONE,IONE,nodes[i]->descU,ZERO,Ue,nodes[i]->u_rank+1,IONE,descUe);
      for(k=br;k>=1;k--)
        if(k!=nodes[i]->U_perm[k-1])
          pswap(nodes[i]->u_rank,Ue,k,IONE,descUe,br,Ue,nodes[i]->U_perm[k-1],IONE,descUe,br);

      /* We expand V into its explicit form V=P*[I;E] */
      locr=numroc_(&bc,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
      locc=numroc_(&nodes[i]->v_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
      Ve=new T[locr*locc]();
      idummy=std::max(1,locr);
      descinit_(descVe,&bc,&nodes[i]->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
      for(k=1;k<=nodes[i]->v_rank;k++)
        if(nodes[i]->myrow==indxg2p_(&k,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow) && nodes[i]->mycol==indxg2p_(&k,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol))
          Ve[(indxg2l_(&k,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol)-1)*locr+indxg2l_(&k,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow)-1]=ONE;
      pgeadd('N',nodes[i]->v_rows,nodes[i]->v_rank,ONE,nodes[i]->V_E,IONE,IONE,nodes[i]->descV,ZERO,Ve,nodes[i]->v_rank+1,IONE,descVe);
      for(k=bc;k>=1;k--)
        if(k!=nodes[i]->V_perm[k-1])
          pswap(nodes[i]->v_rank,Ve,k,IONE,descVe,bc,Ve,nodes[i]->V_perm[k-1],IONE,descVe,bc);

      /* Leaf: U^big=U, V^big=V, stored in explicit form */
      Ubig[i]=Ue;
      Vbig[i]=Ve;
    }
    rowsUbig[i]=br;
    rowsVbig[i]=bc;
  }

  /* Interior nodes */
  for(j=0;j<par_lvl;j++) {
    i=par_nodes[j];

    /* Visit only nodes that are below q */
    if(i<=knode || i>qnode)
      break;

    proc_b=nodes[i]->procs[0];
    proc_g=nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1;

    child1=nodes[i]->children[0];
    child2=nodes[i]->children[1];
    ch1=children1[i];
    ch2=children2[i];

    br=child1->u_rank+child2->u_rank;
    bc=child1->v_rank+child2->v_rank;

    /* We redistribute U_ch1^big and U_ch2^big.
     * First, we communicate number of rows.
     */
    exchangeInteger(&rowsUbig[ch1],child1->procs[0],child1->procs[1],&rowsUbig[ch2],child2->procs[0],child2->procs[1],comm);
    rowsUbig[i]=rowsUbig[ch1]+rowsUbig[ch2];

    /* Redistribute Ubig[ch1] */
    proc_bc=child1->procs[0];
    proc_gc=child1->procs[0]+child1->nprow*child1->npcol-1;
    if(myid>=proc_bc && myid<=proc_gc) {
      /* On child1 grid */
      locr=numroc_(&rowsUbig[ch1],&mb,&child1->myrow,&IZERO,&child1->nprow);
      locc=numroc_(&child1->u_rank,&nb,&child1->mycol,&IZERO,&child1->npcol);
      idummy=std::max(1,locr);
      descinit_(desc_ch1,&rowsUbig[ch1],&child1->u_rank,&mb,&nb,&IZERO,&IZERO,&child1->context,&idummy,&ierr);
    } else
      descset_(desc_ch1,&rowsUbig[ch1],&child1->u_rank,&mb,&nb,&IZERO,&IZERO,&INONE,&INONE);

    if(myid>=proc_b && myid<=proc_g) {
      /* On parent grid */
      locr=numroc_(&rowsUbig[ch1],&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
      locc=numroc_(&child1->u_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
      Xbig1=new T[locr*locc];
      idummy=std::max(1,locr);
      descinit_(desc1,&rowsUbig[ch1],&child1->u_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
    } else
      descset_(desc1,&rowsUbig[ch1],&child1->u_rank,&mb,&nb,&IZERO,&IZERO,&INONE,&INONE);

    /* Ubig[ch1] becomes Xbig1 */
    pgemr2d(rowsUbig[ch1],child1->u_rank,Ubig[ch1],IONE,IONE,desc_ch1,Xbig1,IONE,IONE,desc1,nodes[i]->allctxt);
    delete[] Ubig[ch1];

    /* Redistribute Ubig[ch2] */
    proc_bc=child2->procs[0];
    proc_gc=child2->procs[0]+child2->nprow*child2->npcol-1;
    if(myid>=proc_bc && myid<=proc_gc) {
      /* On child2 grid */
      locr=numroc_(&rowsUbig[ch2],&mb,&child2->myrow,&IZERO,&child2->nprow);
      locc=numroc_(&child2->u_rank,&nb,&child2->mycol,&IZERO,&child2->npcol);
      idummy=std::max(1,locr);
      descinit_(desc_ch2,&rowsUbig[ch2],&child2->u_rank,&mb,&nb,&IZERO,&IZERO,&child2->context,&idummy,&ierr);
    } else
      descset_(desc_ch2,&rowsUbig[ch2],&child2->u_rank,&mb,&nb,&IZERO,&IZERO,&INONE,&INONE);

    if(myid>=proc_b && myid<=proc_g) {
      /* On parent grid */
      locr=numroc_(&rowsUbig[ch2],&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
      locc=numroc_(&child2->u_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
      Xbig2=new T[locr*locc];
      idummy=std::max(1,locr);
      descinit_(desc2,&rowsUbig[ch2],&child2->u_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
    } else
      descset_(desc2,&rowsUbig[ch2],&child2->u_rank,&mb,&nb,&IZERO,&IZERO,&INONE,&INONE);

    /* Ubig[ch2] becomes Xbig2 */
    pgemr2d(rowsUbig[ch2],child2->u_rank,Ubig[ch2],IONE,IONE,desc_ch2,Xbig2,IONE,IONE,desc2,nodes[i]->allctxt);
    delete[] Ubig[ch2];

    if(myid>=proc_b && myid<=proc_g) {
      /* We expand U into its explicit form U=P*[I;E] */
      locr=numroc_(&br,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
      locc=numroc_(&nodes[i]->u_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
      Ue=new T[locr*locc]();
      idummy=std::max(1,locr);
      descinit_(descUe,&br,&nodes[i]->u_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
      for(k=1;k<=nodes[i]->u_rank;k++)
        if(nodes[i]->myrow==indxg2p_(&k,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow) && nodes[i]->mycol==indxg2p_(&k,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol))
          Ue[(indxg2l_(&k,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol)-1)*locr+indxg2l_(&k,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow)-1]=ONE;
      pgeadd('N',nodes[i]->u_rows,nodes[i]->u_rank,ONE,nodes[i]->U_E,IONE,IONE,nodes[i]->descU,ZERO,Ue,nodes[i]->u_rank+1,IONE,descUe);
      for(k=br;k>=1;k--)
        if(k!=nodes[i]->U_perm[k-1])
          pswap(nodes[i]->u_rank,Ue,k,IONE,descUe,br,Ue,nodes[i]->U_perm[k-1],IONE,descUe,br);

      /* We compute [U_ch1^big 0 ; 0 U_ch2^big]*U = [Xbig1*Ue ; Xbig2*Ue] */
      locr=numroc_(&rowsUbig[i],&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
      locc=numroc_(&nodes[i]->u_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
      Ubig[i]=new T[locr*locc]();
      idummy=std::max(1,locr);
      descinit_(desc,&rowsUbig[i],&nodes[i]->u_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
      pgemm('N','N',rowsUbig[ch1],nodes[i]->u_rank,child1->u_rank,ONE,Xbig1,IONE,IONE,desc1,Ue,IONE,IONE,descUe,ZERO,Ubig[i],IONE,IONE,desc);
      pgemm('N','N',rowsUbig[ch2],nodes[i]->u_rank,child2->u_rank,ONE,Xbig2,IONE,IONE,desc2,Ue,child1->u_rank+1,IONE,descUe,ZERO,Ubig[i],rowsUbig[ch1]+1,IONE,desc);

      /* Get rid of Ubig[ch1], Ubig[ch2], and Ue */
      delete[] Xbig1;
      delete[] Xbig2;
      delete[] Ue;
    }

    /* Similarly, we exchange V_ch1^big and V_ch2^big */
    exchangeInteger(&rowsVbig[ch1],child1->procs[0],child1->procs[1],&rowsVbig[ch2],child2->procs[0],child2->procs[1],comm);
    rowsVbig[i]=rowsVbig[ch1]+rowsVbig[ch2];

    /* Redistribute Vbig[ch1] */
    proc_bc=child1->procs[0];
    proc_gc=child1->procs[0]+child1->nprow*child1->npcol-1;
    if(myid>=proc_bc && myid<=proc_gc) {
      /* On child1 grid */
      locr=numroc_(&rowsVbig[ch1],&mb,&child1->myrow,&IZERO,&child1->nprow);
      locc=numroc_(&child1->v_rank,&nb,&child1->mycol,&IZERO,&child1->npcol);
      idummy=std::max(1,locr);
      descinit_(desc_ch1,&rowsVbig[ch1],&child1->v_rank,&mb,&nb,&IZERO,&IZERO,&child1->context,&idummy,&ierr);
    } else
      descset_(desc_ch1,&rowsVbig[ch1],&child1->v_rank,&mb,&nb,&IZERO,&IZERO,&INONE,&INONE);

    if(myid>=proc_b && myid<=proc_g) {
      /* On parent grid */
      locr=numroc_(&rowsVbig[ch1],&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
      locc=numroc_(&child1->v_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
      Xbig1=new T[locr*locc];
      idummy=std::max(1,locr);
      descinit_(desc1,&rowsVbig[ch1],&child1->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
    } else
      descset_(desc1,&rowsVbig[ch1],&child1->v_rank,&mb,&nb,&IZERO,&IZERO,&INONE,&INONE);

    /* Vbig[ch1] becomes Xbig1 */
    pgemr2d(rowsVbig[ch1],child1->v_rank,Vbig[ch1],IONE,IONE,desc_ch1,Xbig1,IONE,IONE,desc1,nodes[i]->allctxt);
    delete[] Vbig[ch1];

    /* Redistribute Vbig[ch2] */
    proc_bc=child2->procs[0];
    proc_gc=child2->procs[0]+child2->nprow*child2->npcol-1;
    if(myid>=proc_bc && myid<=proc_gc) {
      /* On child2 grid */
      locr=numroc_(&rowsVbig[ch2],&mb,&child2->myrow,&IZERO,&child2->nprow);
      locc=numroc_(&child2->v_rank,&nb,&child2->mycol,&IZERO,&child2->npcol);
      idummy=std::max(1,locr);
      descinit_(desc_ch2,&rowsVbig[ch2],&child2->v_rank,&mb,&nb,&IZERO,&IZERO,&child2->context,&idummy,&ierr);
    } else
      descset_(desc_ch2,&rowsVbig[ch2],&child2->v_rank,&mb,&nb,&IZERO,&IZERO,&INONE,&INONE);

    if(myid>=proc_b && myid<=proc_g) {
      /* On parent grid */
      locr=numroc_(&rowsVbig[ch2],&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
      locc=numroc_(&child2->v_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
      Xbig2=new T[locr*locc];
      idummy=std::max(1,locr);
      descinit_(desc2,&rowsVbig[ch2],&child2->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
    } else
      descset_(desc2,&rowsVbig[ch2],&child2->v_rank,&mb,&nb,&IZERO,&IZERO,&INONE,&INONE);

    /* Vbig[ch2] becomes Xbig2 */
    pgemr2d(rowsVbig[ch2],child2->v_rank,Vbig[ch2],IONE,IONE,desc_ch2,Xbig2,IONE,IONE,desc2,nodes[i]->allctxt);
    delete[] Vbig[ch2];

    if(myid>=proc_b && myid<=proc_g) {
      /* We expand V into its explicit form V=P*[I;E] */
      locr=numroc_(&bc,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
      locc=numroc_(&nodes[i]->v_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
      Ve=new T[locr*locc]();
      idummy=std::max(1,locr);
      descinit_(descVe,&bc,&nodes[i]->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
      for(k=1;k<=nodes[i]->v_rank;k++)
        if(nodes[i]->myrow==indxg2p_(&k,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow) && nodes[i]->mycol==indxg2p_(&k,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol))
          Ve[(indxg2l_(&k,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol)-1)*locr+indxg2l_(&k,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow)-1]=ONE;
      pgeadd('N',nodes[i]->v_rows,nodes[i]->v_rank,ONE,nodes[i]->V_E,IONE,IONE,nodes[i]->descV,ZERO,Ve,nodes[i]->v_rank+1,IONE,descVe);
      for(k=bc;k>=1;k--)
        if(k!=nodes[i]->V_perm[k-1])
          pswap(nodes[i]->v_rank,Ve,k,IONE,descVe,bc,Ve,nodes[i]->V_perm[k-1],IONE,descVe,bc);

      /* We compute [V_ch1^big 0 ; 0 V_ch2^big]*V = [Xbig1*Ve ; Xbig2*Ve] */
      locr=numroc_(&rowsVbig[i],&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
      locc=numroc_(&nodes[i]->v_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
      Vbig[i]=new T[locr*locc]();
      idummy=std::max(1,locr);
      descinit_(desc,&rowsVbig[i],&nodes[i]->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
      pgemm('N','N',rowsVbig[ch1],nodes[i]->v_rank,child1->v_rank,ONE,Xbig1,IONE,IONE,desc1,Ve,IONE,IONE,descVe,ZERO,Vbig[i],IONE,IONE,desc);
      pgemm('N','N',rowsVbig[ch2],nodes[i]->v_rank,child2->v_rank,ONE,Xbig2,IONE,IONE,desc2,Ve,child1->v_rank+1,IONE,descVe,ZERO,Vbig[i],rowsVbig[ch1]+1,IONE,desc);

      /* Get rid of Vbig[ch1], Vbig[ch2], and Ve */
      delete[] Xbig1;
      delete[] Xbig2;
      delete[] Ve;
    }

  }

  /* At this point, we have U_q^big and V^q_big onto the q-node grid.
   * We move them to the grid corresponding to the root node.
   */
  i=dim_tree-1;
  ch1=knode;
  ch2=qnode;
  child1=nodes[knode];
  child2=nodes[qnode];
  exchangeInteger(&rowsUbig[ch1],child1->procs[0],child1->procs[1],&rowsUbig[ch2],child2->procs[0],child2->procs[1],comm);
  exchangeInteger(&rowsVbig[ch1],child1->procs[0],child1->procs[1],&rowsVbig[ch2],child2->procs[0],child2->procs[1],comm);

  proc_b=nodes[i]->procs[0];
  proc_g=nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1;

  /* Redistribute Ubig[ch2=q] */
  proc_bc=child2->procs[0];
  proc_gc=child2->procs[0]+child2->nprow*child2->npcol-1;
  if(myid>=proc_bc && myid<=proc_gc) {
    /* On child2 grid */
    locr=numroc_(&rowsUbig[ch2],&mb,&child2->myrow,&IZERO,&child2->nprow);
    locc=numroc_(&child2->u_rank,&nb,&child2->mycol,&IZERO,&child2->npcol);
    idummy=std::max(1,locr);
    descinit_(desc_ch2,&rowsUbig[ch2],&child2->u_rank,&mb,&nb,&IZERO,&IZERO,&child2->context,&idummy,&ierr);
  } else
    descset_(desc_ch2,&rowsUbig[ch2],&child2->u_rank,&mb,&nb,&IZERO,&IZERO,&INONE,&INONE);

  if(myid>=proc_b && myid<=proc_g) {
    /* On parent grid */
    locr=numroc_(&rowsUbig[ch2],&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
    locc=numroc_(&child2->u_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
    Xbig2=new T[locr*locc];
    idummy=std::max(1,locr);
    descinit_(desc2,&rowsUbig[ch2],&child2->u_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
  } else
    descset_(desc2,&rowsUbig[ch2],&child2->u_rank,&mb,&nb,&IZERO,&IZERO,&INONE,&INONE);

  /* Ubig[ch2=q] becomes Xbig2 then replaces Ubig[ch2] */
  pgemr2d(rowsUbig[ch2],child2->u_rank,Ubig[ch2],IONE,IONE,desc_ch2,Xbig2,IONE,IONE,desc2,nodes[i]->allctxt);
  delete[] Ubig[ch2];
  if(myid>=proc_b && myid<=proc_g)
    Ubig[ch2]=Xbig2;
  else
    Ubig[ch2]=NULL;

  /* Redistribute Vbig[ch2=q] */
  proc_bc=child2->procs[0];
  proc_gc=child2->procs[0]+child2->nprow*child2->npcol-1;
  if(myid>=proc_bc && myid<=proc_gc) {
    /* On child2 grid */
    locr=numroc_(&rowsVbig[ch2],&mb,&child2->myrow,&IZERO,&child2->nprow);
    locc=numroc_(&child2->v_rank,&nb,&child2->mycol,&IZERO,&child2->npcol);
    idummy=std::max(1,locr);
    descinit_(desc_ch2,&rowsVbig[ch2],&child2->v_rank,&mb,&nb,&IZERO,&IZERO,&child2->context,&idummy,&ierr);
  } else
    descset_(desc_ch2,&rowsVbig[ch2],&child2->v_rank,&mb,&nb,&IZERO,&IZERO,&INONE,&INONE);

  if(myid>=proc_b && myid<=proc_g) {
    /* On parent grid */
    locr=numroc_(&rowsVbig[ch2],&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
    locc=numroc_(&child2->v_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
    Xbig2=new T[locr*locc];
    idummy=std::max(1,locr);
    descinit_(desc2,&rowsVbig[ch2],&child2->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
  } else
    descset_(desc2,&rowsVbig[ch2],&child2->v_rank,&mb,&nb,&IZERO,&IZERO,&INONE,&INONE);

  /* Vbig[ch2=q] becomes Xbig2 then replaces Vbig[ch2]; */
  pgemr2d(rowsVbig[ch2],child2->v_rank,Vbig[ch2],IONE,IONE,desc_ch2,Xbig2,IONE,IONE,desc2,nodes[i]->allctxt);
  delete[] Vbig[ch2];
  if(myid>=proc_b && myid<=proc_g)
    Vbig[ch2]=Xbig2;
  else
    Vbig[ch2]=NULL;

  /* theta = U_q^big * B_qk.
   * theta: desc_theta, U_q^big: desc2, Bqk: desc
   */
  rtheta=rowsUbig[qnode];
  ctheta=nodes[knode]->v_rank;
  desc_theta=new int[BLACSCTXTSIZE];
  if(myid>=proc_b && myid<=proc_g) {
    /* On parent grid */
    locr=numroc_(&rtheta,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
    locc=numroc_(&ctheta,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
    theta=new T[locr*locc];
    idummy=std::max(1,locr);
    descinit_(desc_theta,&rtheta,&ctheta,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);

    locr=numroc_(&rtheta,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
    idummy=std::max(1,locr);
    descinit_(desc2,&rtheta,&nodes[qnode]->u_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);

    locr=numroc_(&nodes[qnode]->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
    idummy=std::max(1,locr);
    descinit_(desc,&nodes[qnode]->u_rank,&ctheta,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);

    pgemm('N','N',rtheta,ctheta,nodes[qnode]->u_rank,ONE,Ubig[ch2],IONE,IONE,desc2,nodes[dim_tree-1]->B21,IONE,IONE,desc,ZERO,theta,IONE,IONE,desc_theta);
  } else
    descset_(desc_theta,&rtheta,&ctheta,&mb,&nb,&IZERO,&IZERO,&INONE,&INONE);

#if defined(debug)
  if(myid>=proc_b && myid<=proc_g) {
    locr=numroc_(&rtheta,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
    locc=numroc_(&ctheta,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
    printMatrix("theta: "+SSTR(myid),theta,locr,locc,locr);
  }
#endif

  /* Compute psi = Utilda_k^-1 Ltilda_k^-1 Ptilda_k U_k B_kq V_q^big^*.
   * In the sequential case, U_k * B_kq is computed using the structured
   * product (apply_generator). Here U_k and B_kq are on different grids.
   * To reduce communications, we proceed as follows:
   *  1/ On the k-node grid, compute Ltilda_k^-1 U_k (after computing an
   *     explicit version of U_k, Ue). The result is in Ue, descUe.
   *  2/ Move the result to the root node grid using variables tmp and desc
   *  3/ Reassign Ue=tmp, descUe=desc.
   *  4/ Compute tmp=Ue*B_kq.
   *  5/ Compute psi=tmp*V_q^big.
   */
  proc_bc=child1->procs[0];
  proc_gc=child1->procs[0]+child1->nprow*child1->npcol-1;
  br=nodes[knode]->u_rank+nodes[knode]->u_rows;
  if(myid>=proc_bc && myid<=proc_gc) {
    /* On the k-node grid */
    if(proc_bc==proc_gc) {
      /* If k is mapped on one process, we don't have descriptors. We assign them here. */
      locr=numroc_(&br,&mb,&nodes[ch1]->myrow,&IZERO,&nodes[ch1]->nprow);
      idummy=std::max(1,locr);
      nodes[ch1]->descL=new int[BLACSCTXTSIZE];
      descinit_(nodes[ch1]->descL,&br,&br,&mb,&nb,&IZERO,&IZERO,&nodes[ch1]->context,&idummy,&ierr);

      locr=numroc_(&nodes[ch1]->u_rows,&mb,&nodes[ch1]->myrow,&IZERO,&nodes[ch1]->nprow);
      idummy=std::max(1,locr);
      nodes[ch1]->descU=new int[BLACSCTXTSIZE];
      descinit_(nodes[ch1]->descU,&nodes[ch1]->u_rows,&nodes[ch1]->u_rank,&mb,&nb,&IZERO,&IZERO,&nodes[ch1]->context,&idummy,&ierr);
    }

    /* We expand U into its explicit form U=P*[I;E] */
    locr=numroc_(&br,&mb,&nodes[ch1]->myrow,&IZERO,&nodes[ch1]->nprow);
    locc=numroc_(&nodes[ch1]->u_rank,&nb,&nodes[ch1]->mycol,&IZERO,&nodes[ch1]->npcol);
    Ue=new T[locr*locc]();
    idummy=std::max(1,locr);
    descinit_(descUe,&br,&nodes[ch1]->u_rank,&mb,&nb,&IZERO,&IZERO,&nodes[ch1]->context,&idummy,&ierr);
    for(k=1;k<=nodes[ch1]->u_rank;k++)
      if(nodes[ch1]->myrow==indxg2p_(&k,&mb,&nodes[ch1]->myrow,&IZERO,&nodes[ch1]->nprow) && nodes[ch1]->mycol==indxg2p_(&k,&nb,&nodes[ch1]->mycol,&IZERO,&nodes[ch1]->npcol))
        Ue[(indxg2l_(&k,&nb,&nodes[ch1]->mycol,&IZERO,&nodes[ch1]->npcol)-1)*locr+indxg2l_(&k,&mb,&nodes[ch1]->myrow,&IZERO,&nodes[ch1]->nprow)-1]=ONE;
    pgeadd('N',nodes[ch1]->u_rows,nodes[ch1]->u_rank,ONE,nodes[ch1]->U_E,IONE,IONE,nodes[ch1]->descU,ZERO,Ue,nodes[ch1]->u_rank+1,IONE,descUe);
    for(k=br;k>=1;k--)
      if(k!=nodes[ch1]->U_perm[k-1])
        pswap(nodes[ch1]->u_rank,Ue,k,IONE,descUe,br,Ue,nodes[ch1]->U_perm[k-1],IONE,descUe,br);

    /* 1/ Ltilda_k \ U_k */
    pgetrs('N',br,nodes[ch1]->u_rank,nodes[ch1]->L,IONE,IONE,nodes[ch1]->descL,nodes[ch1]->piv,Ue,IONE,IONE,descUe);

  } else {
    descset_(descUe,&br,&nodes[ch1]->u_rank,&mb,&nb,&IZERO,&IZERO,&INONE,&INONE);
  }

  /* 2/ Move Ltilda_k \ U_k onto the root-node grid */
  if(myid>=proc_b && myid<=proc_g) {
    /* On parent grid */
    locr=numroc_(&br,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
    locc=numroc_(&nodes[ch1]->u_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
    tmp=new T[locr*locc];
    idummy=std::max(1,locr);
    descinit_(desc,&br,&nodes[ch1]->u_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
  } else {
    descset_(desc,&br,&nodes[ch1]->u_rank,&mb,&nb,&IZERO,&IZERO,&INONE,&INONE);
  }

  pgemr2d(br,nodes[ch1]->u_rank,Ue,IONE,IONE,descUe,tmp,IONE,IONE,desc,nodes[i]->allctxt);

  /* 3/ Rename (tmp,desc) to (Ue,descUe) */
  if(myid>=proc_bc && myid<=proc_gc) {
    delete[] Ue;
  }
  if(myid>=proc_b && myid<=proc_g) {
    Ue=tmp;
  } else {
    Ue=NULL;
  }
  copyDesc(desc,descUe);

  desc_psi=new int[BLACSCTXTSIZE];
  rpsi=br;
  cpsi=rowsVbig[qnode];
  if(myid>=proc_b && myid<=proc_g) {
    /* 4/ Compute tmp=Ue*B_kq */
    locr=numroc_(&br,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
    locc=numroc_(&nodes[ch2]->v_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
    tmp=new T[locr*locc];
    idummy=std::max(1,locr);
    descinit_(desc,&br,&nodes[ch2]->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);

    locr=numroc_(&nodes[ch1]->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
    idummy=std::max(1,locr);
    descinit_(desc2,&nodes[ch1]->u_rank,&nodes[ch2]->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);

    pgemm('N','N',br,nodes[ch2]->v_rank,nodes[ch1]->u_rank,ONE,Ue,IONE,IONE,descUe,nodes[i]->B12,IONE,IONE,desc2,ZERO,tmp,IONE,IONE,desc);

    /* 5/ Compute psi=tmp*V_q^big
     * psi: desc_psi, tmp: desc, V_q^big: desc2 */
    locr=numroc_(&rpsi,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
    locc=numroc_(&cpsi,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
    psi=new T[locr*locc];
    idummy=std::max(1,locr);
    descinit_(desc_psi,&rpsi,&cpsi,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);

    locr=numroc_(&rowsVbig[qnode],&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
    idummy=std::max(1,locr);
    descinit_(desc2,&rowsVbig[qnode],&nodes[qnode]->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);

    pgemm('N','C',rpsi,cpsi,nodes[ch2]->v_rank,ONE,tmp,IONE,IONE,desc,Vbig[qnode],IONE,IONE,desc2,ZERO,psi,IONE,IONE,desc_psi);

  } else {
    descset_(desc_psi,&rpsi,&cpsi,&mb,&nb,&IZERO,&IZERO,&INONE,&INONE);

  }

#if defined(debug)
  if(myid>=proc_b && myid<=proc_g) {
    locr=numroc_(&rpsi,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
    locc=numroc_(&cpsi,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
    printMatrix("psi: "+SSTR(myid),psi,locr,locc,locr);
  }
#endif

  /* The last thing to do is to move nodes[knode]->Vt from the k-node grid
   * to the root-node grid. It is used only for Schur complement products,
   * not for partial triangular solves, so it is safe to modify it.
   */
  proc_bc=child1->procs[0];
  proc_gc=child1->procs[0]+child1->nprow*child1->npcol-1;
  br=child1->u_rows+child1->u_rank;
  if(myid>=proc_bc && myid<=proc_gc) {
    /* On child1/knode grid */
    locr=numroc_(&br,&mb,&child1->myrow,&IZERO,&child1->nprow);
    locc=numroc_(&child1->v_rank,&nb,&child1->mycol,&IZERO,&child1->npcol);
    idummy=std::max(1,locr);
    descinit_(desc_ch1,&br,&child1->v_rank,&mb,&nb,&IZERO,&IZERO,&child1->context,&idummy,&ierr);
  } else
    descset_(desc_ch1,&br,&child1->v_rank,&mb,&nb,&IZERO,&IZERO,&INONE,&INONE);

  if(myid>=proc_b && myid<=proc_g) {
    /* On parent grid */
    locr=numroc_(&br,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
    locc=numroc_(&child1->v_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
    tmp=new T[locr*locc];
    idummy=std::max(1,locr);
    descinit_(desc1,&br,&child1->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
  } else
    descset_(desc1,&br,&child1->v_rank,&mb,&nb,&IZERO,&IZERO,&INONE,&INONE);

  pgemr2d(br,child1->v_rank,child1->Vt,IONE,IONE,desc_ch1,tmp,IONE,IONE,desc1,nodes[i]->allctxt);
  if(myid>=proc_bc && myid<=proc_gc)
    delete[] child1->Vt;
  if(myid>=proc_b && myid<=proc_g)
    child1->Vt=tmp;
  else
    child1->Vt=NULL;
  child1->descVt=new int[BLACSCTXTSIZE];
  copyDesc(desc1,child1->descVt);

#if defined(debug)
  if(myid>=proc_b && myid<=proc_g) {
    printMatrix("Vt: "+SSTR(myid),child1->Vt,locr,locc,locr);
  }
#endif

  delete[] Ubig[qnode];
  delete[] Vbig[qnode];

  delete[] rowsUbig;
  delete[] rowsVbig;

}

#endif // HSS_PAR_GENERATESCHUR_HPP
