/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef STRUMPACK_DENSE_PACKAGE_H
#define STRUMPACK_DENSE_PACKAGE_H

#include <stdlib.h>
#include <math.h>
#include "constants.h"
#include "mpi.h"
#include "blacs.h"
#include "blas.h"
#include "lapack.h"
#include "scalapack.h"

/* C version of the Partition_tree class */
typedef struct Partition_tree_C {
  int size;
  struct Partition_tree_C* c[2];
} Partition_tree_C;

/* Structures StrumpackDensePackage_C_* mimic the StrumpackDensePackage class;
 * We explicitly define four versions: double, float, dcomplex, scomplex.
 * We could use macros but it would make using different arithmetic
 * simultaneously complicated.
 */

/*****************************************************************************/
/* Real double precision */

typedef struct StrumpackDensePackage_C_double {
  /* User options */
  int  use_HSS;            /* Use HSS compression/factorization/solve or classical LU? */
  double tol_HSS;          /* Tolerance for HSS compression */
  int  min_rand_HSS;       /* Minimum number of random vectors for HSS compression */
  int  max_rand_HSS;       /* Maximum number of random vectors for HSS compression */
  int  lim_rand_HSS;       /* Criterion of increasing the number of random vectors */
  int  inc_rand_HSS;       /* Number of extra random vectors at each increase */
  int  block_HSS;          /* Block size for the leaves of the HSS tree */
  int  levels_HSS;         /* Number of levels in the HSS tree */
  int  split_HSS;          /* Constrained partitioning of the root node */
  double tol_IR;           /* Tolerance for iterative refinement */
  int  steps_IR;           /* Number of steps of iterative refinement */
  int  fast_IR;            /* Matvecs with HSS(A) or A in iterative refinement? */
  int  verbose;            /* Verbose or silent? */
  int  access_is_global;   /* User-given access routine: requires global synchronization? */
  int  sched_strat;        /* Scheduling strategy: 1 or 2 */
  void *obj;               /* Handle to user data/object, for strategy 2 */
  Partition_tree_C *utree; /* User-given clustering */

  /* Handle to the C++ class */
  void *SDP;
} StrumpackDensePackage_C_double;

/* User routines */
void SDP_C_double_init(StrumpackDensePackage_C_double *SDP_C, MPI_Comm comm);
void SDP_C_double_destroy(StrumpackDensePackage_C_double *SDP_C);
void SDP_C_double_compress_A(StrumpackDensePackage_C_double *SDP_C, double *A, int *descA);
void SDP_C_double_compress_ARS(StrumpackDensePackage_C_double *SDP_C, double* A, int *descA, double *Rr, double *Rc, double* Sr, double* Sc, int *descRS);
void SDP_C_double_compress_mtxfree(StrumpackDensePackage_C_double *SDP_C, double *Rr, double*Rc, double*Sr, double* Sc, int *descRS, void (*elements)(void *, int *, int *, double *, int *));
double SDP_C_double_check_compression(StrumpackDensePackage_C_double *SDP_C, double *A, int *descA);
void SDP_C_double_factor(StrumpackDensePackage_C_double *SDP_C, double *A, int *descA);
void SDP_C_double_partially_factor(StrumpackDensePackage_C_double *SDP_C, double* A, int *descA, int nfact);
void SDP_C_double_compute_schur(StrumpackDensePackage_C_double *SDP_C);
void SDP_C_double_reduce_RHS(StrumpackDensePackage_C_double *SDP_C, double *X, int *descX, double* B, int *descB);
void SDP_C_double_schur_product(StrumpackDensePackage_C_double *SDP_C, char Trans, double alpha, double *X, int *descX, double beta, double *B, int *descB);
void SDP_C_double_expand_solution(StrumpackDensePackage_C_double *SDP_C, double *X, int *descX, double *B, int *descB);
void SDP_C_double_solve(StrumpackDensePackage_C_double *SDP_C, double *X, int *descX, double *B, int *descB);
double SDP_C_double_check_solution(StrumpackDensePackage_C_double *SDP_C, double *A, int *descA, double *X, int *descX, double *B, int *descB);
double SDP_C_double_refine(StrumpackDensePackage_C_double *SDP_C, double *A, int *descA, double *X, int *descX, double *B, int *descB);
void SDP_C_double_product(StrumpackDensePackage_C_double *SDP_C, char Trans, double alpha, double *A, int *descA, double *B, int *descB, double beta, double *C, int *descC);
void SDP_C_double_extract(StrumpackDensePackage_C_double *SDP_C, double *A, int *descA, double *B, int *descB, int *I, int nI, int *J, int nJ);
void SDP_C_double_extract_schur(StrumpackDensePackage_C_double *SDP_C, double *B, int *descB, int *I, int nI, int *J, int nJ);
void SDP_C_double_print_statistics(StrumpackDensePackage_C_double *SDP_C);

/* Utility routine */
void SDP_C_double_copy_fields(StrumpackDensePackage_C_double *SDP_C);

/*****************************************************************************/
/* Real single precision */

typedef struct StrumpackDensePackage_C_float {
  /* User options */
  int  use_HSS;            /* Use HSS compression/factorization/solve or classical LU? */
  float tol_HSS;           /* Tolerance for HSS compression */
  int  min_rand_HSS;       /* Minimum number of random vectors for HSS compression */
  int  max_rand_HSS;       /* Maximum number of random vectors for HSS compression */
  int  lim_rand_HSS;       /* Criterion of increasing the number of random vectors */
  int  inc_rand_HSS;       /* Number of extra random vectors at each increase */
  int  block_HSS;          /* Block size for the leaves of the HSS tree */
  int  levels_HSS;         /* Number of levels in the HSS tree */
  int  split_HSS;          /* Constrained partitioning of the root node */
  float tol_IR;            /* Tolerance for iterative refinement */
  int  steps_IR;           /* Number of steps of iterative refinement */
  int  fast_IR;            /* Matvecs with HSS(A) or A in iterative refinement? */
  int  verbose;            /* Verbose or silent? */
  int  access_is_global;   /* User-given access routine: requires global synchronization? */
  int  sched_strat;        /* Scheduling strategy: 1 or 2 */
  void *obj;               /* Handle to user data/object, for strategy 2 */
  Partition_tree_C *utree; /* User-given clustering */

  /* Handle to the C++ class */
  void *SDP;
} StrumpackDensePackage_C_float;

/* User routines */
void SDP_C_float_init(StrumpackDensePackage_C_float *SDP_C, MPI_Comm comm);
void SDP_C_float_destroy(StrumpackDensePackage_C_float *SDP_C);
void SDP_C_float_compress_A(StrumpackDensePackage_C_float *SDP_C, float *A, int *descA);
void SDP_C_float_compress_ARS(StrumpackDensePackage_C_float *SDP_C, float* A, int *descA, float *Rr, float *Rc, float* Sr, float* Sc, int *descRS);
void SDP_C_float_compress_mtxfree(StrumpackDensePackage_C_float *SDP_C, float *Rr, float*Rc, float*Sr, float* Sc, int *descRS, void (*elements)(void *, int *, int *, float *, int *));
float SDP_C_float_check_compression(StrumpackDensePackage_C_float *SDP_C, float *A, int *descA);
void SDP_C_float_factor(StrumpackDensePackage_C_float *SDP_C, float *A, int *descA);
void SDP_C_float_partially_factor(StrumpackDensePackage_C_float *SDP_C, float* A, int *descA, int nfact);
void SDP_C_float_compute_schur(StrumpackDensePackage_C_float *SDP_C);
void SDP_C_float_reduce_RHS(StrumpackDensePackage_C_float *SDP_C, float *X, int *descX, float* B, int *descB);
void SDP_C_float_schur_product(StrumpackDensePackage_C_float *SDP_C, char Trans, float alpha, float *X, int *descX, float beta, float *B, int *descB);
void SDP_C_float_expand_solution(StrumpackDensePackage_C_float *SDP_C, float *X, int *descX, float *B, int *descB);
void SDP_C_float_solve(StrumpackDensePackage_C_float *SDP_C, float *X, int *descX, float *B, int *descB);
float SDP_C_float_check_solution(StrumpackDensePackage_C_float *SDP_C, float *A, int *descA, float *X, int *descX, float *B, int *descB);
float SDP_C_float_refine(StrumpackDensePackage_C_float *SDP_C, float *A, int *descA, float *X, int *descX, float *B, int *descB);
void SDP_C_float_product(StrumpackDensePackage_C_float *SDP_C, char Trans, float alpha, float *A, int *descA, float *B, int *descB, float beta, float *C, int *descC);
void SDP_C_float_extract(StrumpackDensePackage_C_float *SDP_C, float *A, int *descA, float *B, int *descB, int *I, int nI, int *J, int nJ);
void SDP_C_float_extract_schur(StrumpackDensePackage_C_float *SDP_C, float *B, int *descB, int *I, int nI, int *J, int nJ);
void SDP_C_float_print_statistics(StrumpackDensePackage_C_float *SDP_C);

/* Utility routine */
void SDP_C_float_copy_fields(StrumpackDensePackage_C_float *SDP_C);

/*****************************************************************************/
/* Complex double precision */

typedef struct StrumpackDensePackage_C_dcomplex {
  /* User options */
  int  use_HSS;            /* Use HSS compression/factorization/solve or classical LU? */
  double tol_HSS;          /* Tolerance for HSS compression */
  int  min_rand_HSS;       /* Minimum number of random vectors for HSS compression */
  int  max_rand_HSS;       /* Maximum number of random vectors for HSS compression */
  int  lim_rand_HSS;       /* Criterion of increasing the number of random vectors */
  int  inc_rand_HSS;       /* Number of extra random vectors at each increase */
  int  block_HSS;          /* Block size for the leaves of the HSS tree */
  int  levels_HSS;         /* Number of levels in the HSS tree */
  int  split_HSS;          /* Constrained partitioning of the root node */
  double tol_IR;           /* Tolerance for iterative refinement */
  int  steps_IR;           /* Number of steps of iterative refinement */
  int  fast_IR;            /* Matvecs with HSS(A) or A in iterative refinement? */
  int  verbose;            /* Verbose or silent? */
  int  access_is_global;   /* User-given access routine: requires global synchronization? */
  int  sched_strat;        /* Scheduling strategy: 1 or 2 */
  void *obj;               /* Handle to user data/object, for strategy 2 */
  Partition_tree_C *utree; /* User-given clustering */

  /* Handle to the C++ class */
  void *SDP;
} StrumpackDensePackage_C_dcomplex;

/* User routines */
void SDP_C_dcomplex_init(StrumpackDensePackage_C_dcomplex *SDP_C, MPI_Comm comm);
void SDP_C_dcomplex_destroy(StrumpackDensePackage_C_dcomplex *SDP_C);
void SDP_C_dcomplex_compress_A(StrumpackDensePackage_C_dcomplex *SDP_C, dcomplex *A, int *descA);
void SDP_C_dcomplex_compress_ARS(StrumpackDensePackage_C_dcomplex *SDP_C, dcomplex* A, int *descA, dcomplex *Rr, dcomplex *Rc, dcomplex* Sr, dcomplex* Sc, int *descRS);
void SDP_C_dcomplex_compress_mtxfree(StrumpackDensePackage_C_dcomplex *SDP_C, dcomplex *Rr, dcomplex*Rc, dcomplex*Sr, dcomplex* Sc, int *descRS, void (*elements)(void *, int *, int *, dcomplex *, int *));
double SDP_C_dcomplex_check_compression(StrumpackDensePackage_C_dcomplex *SDP_C, dcomplex *A, int *descA);
void SDP_C_dcomplex_factor(StrumpackDensePackage_C_dcomplex *SDP_C, dcomplex *A, int *descA);
void SDP_C_dcomplex_partially_factor(StrumpackDensePackage_C_dcomplex *SDP_C, dcomplex* A, int *descA, int nfact);
void SDP_C_dcomplex_compute_schur(StrumpackDensePackage_C_dcomplex *SDP_C);
void SDP_C_dcomplex_reduce_RHS(StrumpackDensePackage_C_dcomplex *SDP_C, dcomplex *X, int *descX, dcomplex* B, int *descB);
void SDP_C_dcomplex_schur_product(StrumpackDensePackage_C_dcomplex *SDP_C, char Trans, dcomplex alpha, dcomplex *X, int *descX, dcomplex beta, dcomplex *B, int *descB);
void SDP_C_dcomplex_expand_solution(StrumpackDensePackage_C_dcomplex *SDP_C, dcomplex *X, int *descX, dcomplex *B, int *descB);
void SDP_C_dcomplex_solve(StrumpackDensePackage_C_dcomplex *SDP_C, dcomplex *X, int *descX, dcomplex *B, int *descB);
double SDP_C_dcomplex_check_solution(StrumpackDensePackage_C_dcomplex *SDP_C, dcomplex *A, int *descA, dcomplex *X, int *descX, dcomplex *B, int *descB);
double SDP_C_dcomplex_refine(StrumpackDensePackage_C_dcomplex *SDP_C, dcomplex *A, int *descA, dcomplex *X, int *descX, dcomplex *B, int *descB);
void SDP_C_dcomplex_product(StrumpackDensePackage_C_dcomplex *SDP_C, char Trans, dcomplex alpha, dcomplex *A, int *descA, dcomplex *B, int *descB, dcomplex beta, dcomplex *C, int *descC);
void SDP_C_dcomplex_extract(StrumpackDensePackage_C_dcomplex *SDP_C, dcomplex *A, int *descA, dcomplex *B, int *descB, int *I, int nI, int *J, int nJ);
void SDP_C_dcomplex_extract_schur(StrumpackDensePackage_C_dcomplex *SDP_C, dcomplex *B, int *descB, int *I, int nI, int *J, int nJ);
void SDP_C_dcomplex_print_statistics(StrumpackDensePackage_C_dcomplex *SDP_C);

/* Utility routine */
void SDP_C_dcomplex_copy_fields(StrumpackDensePackage_C_dcomplex *SDP_C);

/*****************************************************************************/
/* Complex single precision */

typedef struct StrumpackDensePackage_C_scomplex {
  /* User options */
  int  use_HSS;            /* Use HSS compression/factorization/solve or classical LU? */
  float tol_HSS;           /* Tolerance for HSS compression */
  int  min_rand_HSS;       /* Minimum number of random vectors for HSS compression */
  int  max_rand_HSS;       /* Maximum number of random vectors for HSS compression */
  int  lim_rand_HSS;       /* Criterion of increasing the number of random vectors */
  int  inc_rand_HSS;       /* Number of extra random vectors at each increase */
  int  block_HSS;          /* Block size for the leaves of the HSS tree */
  int  levels_HSS;         /* Number of levels in the HSS tree */
  int  split_HSS;          /* Constrained partitioning of the root node */
  float tol_IR;            /* Tolerance for iterative refinement */
  int  steps_IR;           /* Number of steps of iterative refinement */
  int  fast_IR;            /* Matvecs with HSS(A) or A in iterative refinement? */
  int  verbose;            /* Verbose or silent? */
  int  access_is_global;   /* User-given access routine: requires global synchronization? */
  int  sched_strat;        /* Scheduling strategy: 1 or 2 */
  void *obj;               /* Handle to user data/object, for strategy 2 */
  Partition_tree_C *utree; /* User-given clustering */

  /* Handle to the C++ class */
  void *SDP;
} StrumpackDensePackage_C_scomplex;

/* User routines */
void SDP_C_scomplex_init(StrumpackDensePackage_C_scomplex *SDP_C, MPI_Comm comm);
void SDP_C_scomplex_destroy(StrumpackDensePackage_C_scomplex *SDP_C);
void SDP_C_scomplex_compress_A(StrumpackDensePackage_C_scomplex *SDP_C, scomplex *A, int *descA);
void SDP_C_scomplex_compress_ARS(StrumpackDensePackage_C_scomplex *SDP_C, scomplex* A, int *descA, scomplex *Rr, scomplex *Rc, scomplex* Sr, scomplex* Sc, int *descRS);
void SDP_C_scomplex_compress_mtxfree(StrumpackDensePackage_C_scomplex *SDP_C, scomplex *Rr, scomplex*Rc, scomplex*Sr, scomplex* Sc, int *descRS, void (*elements)(void *, int *, int *, scomplex *, int *));
float SDP_C_scomplex_check_compression(StrumpackDensePackage_C_scomplex *SDP_C, scomplex *A, int *descA);
void SDP_C_scomplex_factor(StrumpackDensePackage_C_scomplex *SDP_C, scomplex *A, int *descA);
void SDP_C_scomplex_partially_factor(StrumpackDensePackage_C_scomplex *SDP_C, scomplex* A, int *descA, int nfact);
void SDP_C_scomplex_compute_schur(StrumpackDensePackage_C_scomplex *SDP_C);
void SDP_C_scomplex_reduce_RHS(StrumpackDensePackage_C_scomplex *SDP_C, scomplex *X, int *descX, scomplex* B, int *descB);
void SDP_C_scomplex_schur_product(StrumpackDensePackage_C_scomplex *SDP_C, char Trans, scomplex alpha, scomplex *X, int *descX, scomplex beta, scomplex *B, int *descB);
void SDP_C_scomplex_expand_solution(StrumpackDensePackage_C_scomplex *SDP_C, scomplex *X, int *descX, scomplex *B, int *descB);
void SDP_C_scomplex_solve(StrumpackDensePackage_C_scomplex *SDP_C, scomplex *X, int *descX, scomplex *B, int *descB);
float SDP_C_scomplex_check_solution(StrumpackDensePackage_C_scomplex *SDP_C, scomplex *A, int *descA, scomplex *X, int *descX, scomplex *B, int *descB);
float SDP_C_scomplex_refine(StrumpackDensePackage_C_scomplex *SDP_C, scomplex *A, int *descA, scomplex *X, int *descX, scomplex *B, int *descB);
void SDP_C_scomplex_product(StrumpackDensePackage_C_scomplex *SDP_C, char Trans, scomplex alpha, scomplex *A, int *descA, scomplex *B, int *descB, scomplex beta, scomplex *C, int *descC);
void SDP_C_scomplex_extract(StrumpackDensePackage_C_scomplex *SDP_C, scomplex *A, int *descA, scomplex *B, int *descB, int *I, int nI, int *J, int nJ);
void SDP_C_scomplex_extract_schur(StrumpackDensePackage_C_scomplex *SDP_C, scomplex *B, int *descB, int *I, int nI, int *J, int nJ);
void SDP_C_scomplex_print_statistics(StrumpackDensePackage_C_scomplex *SDP_C);

/* Utility routine */
void SDP_C_scomplex_copy_fields(StrumpackDensePackage_C_scomplex *SDP_C);

#endif // STRUMPACK_DENSE_PACKAGE_H
