/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */
#ifndef FRONTAL_MATRIX_DENSE_HPP
#define FRONTAL_MATRIX_DENSE_HPP

#include <iostream>
#include <fstream>
#include <algorithm>
#include <cmath>
#include <random>
#include "blas_lapack_wrapper.hpp"
#include "blas_lapack_omp_task.hpp"
#include "CompressedSparseMatrix.hpp"
#include "MatrixReordering.hpp"
#include "TaskTimer.hpp"

namespace strumpack {

  template<typename scalar_t,typename real_t,typename integer_t> class FrontalMatrixDense
    : public FrontalMatrix<scalar_t,real_t,integer_t> {
  public:
    scalar_t* F11;
    scalar_t* F12;
    scalar_t* F21;
    scalar_t* F22;
    int* piv; // this should be a regular int because it is passed to BLAS

    FrontalMatrixDense(CompressedSparseMatrix<scalar_t,real_t,integer_t>* _A,
		       integer_t _sep, integer_t _sep_begin, integer_t _sep_end,
		       integer_t _dim_upd, integer_t* _upd);
    ~FrontalMatrixDense();
    void release_work_memory();
    void extend_add(FrontalMatrixDense<scalar_t,real_t,integer_t>* ch, int task_depth);
    void UUtxR(integer_t cols, scalar_t* R, scalar_t* Srow, scalar_t* Scol, FrontalMatrix<scalar_t,real_t,integer_t>* pa, int task_depth);
    void multifrontal_factorization(HSSStrategy<real_t>& strat, int etree_level=0, int task_depth=0);
    void forward_multifrontal_solve(scalar_t* b, scalar_t* wmem, int etree_level=0, int task_depth=0);
    void backward_multifrontal_solve(scalar_t* y, scalar_t* wmem, int etree_level=0, int task_depth=0);
    void get_submatrix(integer_t* rows, integer_t nr_rows, integer_t* cols, integer_t nr_cols,
		       scalar_t* values, integer_t ld, int task_depth);
    std::string type() { return "FrontalMatrixDense"; }

  private:
    FrontalMatrixDense(const FrontalMatrixDense&) = delete;
    FrontalMatrixDense& operator=(FrontalMatrixDense const&) = delete;
    void multifrontal_factorization_phase1(HSSStrategy<real_t>& strat, int etree_level, int task_depth, bool tasked);
    void multifrontal_factorization_phase2(HSSStrategy<real_t>& strat, int etree_level, int task_depth, bool tasked);
    void forward_multifrontal_solve_phase1(scalar_t* b, scalar_t* wmem, int etree_level, int task_depth, bool tasked);
    void forward_multifrontal_solve_phase2(scalar_t* b, scalar_t* wmem, int etree_level, int task_depth, bool tasked);
    void backward_multifrontal_solve_phase1(scalar_t* y, scalar_t* wmem, int etree_level, int task_depth, bool tasked);
    void backward_multifrontal_solve_phase2(scalar_t* y, scalar_t* wmem, int etree_level, int task_depth, bool tasked);
#if defined(dbgHSS)
    friend class FrontalMatrixHSS<scalar_t,real_t,integer_t>;
#endif
  };

  template<typename scalar_t,typename real_t,typename integer_t>
  FrontalMatrixDense<scalar_t,real_t,integer_t>::FrontalMatrixDense
  (CompressedSparseMatrix<scalar_t,real_t,integer_t>* _A, integer_t _sep,
   integer_t _sep_begin, integer_t _sep_end, integer_t _dim_upd, integer_t* _upd)
    : FrontalMatrix<scalar_t,real_t,integer_t>(_A, NULL, NULL, _sep, _sep_begin, _sep_end, _dim_upd, _upd),
    F11(NULL), F12(NULL), F21(NULL), F22(NULL), piv(NULL) {
  }

  template<typename scalar_t,typename real_t,typename integer_t>
  FrontalMatrixDense<scalar_t,real_t,integer_t>::~FrontalMatrixDense() {
    delete[] F11;
    delete[] piv;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDense<scalar_t,real_t,integer_t>::release_work_memory() {
    delete[] F22; F22 = NULL;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDense<scalar_t,real_t,integer_t>::extend_add
  (FrontalMatrixDense<scalar_t,real_t,integer_t>* ch, int task_depth) {
    bool tasked = task_depth < params::task_recursion_cutoff_level;
    integer_t dim_sep = this->dim_sep;  integer_t dim_upd = this->dim_upd;
    integer_t ch_dim_upd = ch->dim_upd;
    std::vector<integer_t> parent_ind;
    integer_t upd_rows_to_sep;
    this->compute_parent_indices(ch, parent_ind, upd_rows_to_sep);

    if (tasked) {
      integer_t loop_tasks = std::max(params::num_threads / (task_depth+1), 1);
      integer_t B = std::max(ch_dim_upd / loop_tasks, integer_t(1));
      for (integer_t task=0; task<std::ceil(ch_dim_upd/float(B)); task++)
#pragma omp task default(shared) firstprivate(task)
	{
	  for (integer_t col=task*B; col<std::min((task+1)*B,ch_dim_upd); col++) {
	    integer_t pc = parent_ind[col];
	    if (pc < dim_sep) {
	      for (integer_t row=0; row<upd_rows_to_sep; row++)
		F11[parent_ind[row] + pc*dim_sep] += ch->F22[row + col*ch_dim_upd];
	      for (integer_t row=upd_rows_to_sep; row<ch_dim_upd; row++)
		F21[parent_ind[row] - dim_sep + pc*dim_upd] += ch->F22[row + col*ch_dim_upd];
	    } else {
	      for (integer_t row=0; row<upd_rows_to_sep; row++)
		F12[parent_ind[row] + (pc-dim_sep)*dim_sep] += ch->F22[row + col*ch_dim_upd];
	      for (integer_t row=upd_rows_to_sep; row<ch_dim_upd; row++)
		F22[parent_ind[row] - dim_sep + (pc - dim_sep)*dim_upd] += ch->F22[row + col*ch_dim_upd];
	    }
	  }
	}
#pragma omp taskwait
    } else {
      for (integer_t col=0; col<ch_dim_upd; col++) {
	integer_t pc = parent_ind[col];
	if (pc < dim_sep) {
	  for (integer_t row=0; row<upd_rows_to_sep; row++)
	    F11[parent_ind[row] + pc*dim_sep] += ch->F22[row + col*ch_dim_upd];
	  for (integer_t row=upd_rows_to_sep; row<ch_dim_upd; row++)
	    F21[parent_ind[row] - dim_sep + pc*dim_upd] += ch->F22[row + col*ch_dim_upd];
	} else {
	  for (integer_t row=0; row<upd_rows_to_sep; row++)
	    F12[parent_ind[row] + (pc-dim_sep)*dim_sep] += ch->F22[row + col*ch_dim_upd];
	  for (integer_t row=upd_rows_to_sep; row<ch_dim_upd; row++)
	    F22[parent_ind[row] - dim_sep + (pc - dim_sep)*dim_upd] += ch->F22[row + col*ch_dim_upd];
	}
      }
    }
    STRUMPACK_FLOPS((is_complex<scalar_t>()?2:1)*static_cast<long long int>(ch_dim_upd*ch_dim_upd));
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDense<scalar_t,real_t,integer_t>::UUtxR
  (integer_t cols, scalar_t* R, scalar_t* Srow, scalar_t* Scol,
   FrontalMatrix<scalar_t,real_t,integer_t>* pa, int task_depth) {
    if (this->dim_upd == 0) return;
    auto Rtmp = new scalar_t[2*cols*this->dim_upd];
    auto Stmp = &Rtmp[cols*this->dim_upd];
    this->extract_child_rows(cols, R, Rtmp, pa, task_depth);
    gemm_omp_task('N', 'N', this->dim_upd, cols, this->dim_upd, scalar_t(1.), F22, this->dim_upd,
		  Rtmp, this->dim_upd, scalar_t(0.), Stmp, this->dim_upd, task_depth);
    this->skinny_extend_add(cols, Stmp, pa, Srow, task_depth);
    gemm_omp_task('C', 'N', this->dim_upd, cols, this->dim_upd, scalar_t(1.), F22, this->dim_upd,
		  Rtmp, this->dim_upd, scalar_t(0.), Stmp, this->dim_upd, task_depth);
    this->skinny_extend_add(cols, Stmp, pa, Scol, task_depth);
    delete[] Rtmp;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDense<scalar_t,real_t,integer_t>::multifrontal_factorization
  (HSSStrategy<real_t>& strat, int etree_level, int task_depth) {
    if (task_depth == 0) {
#pragma omp parallel if(!omp_in_parallel())
#pragma omp single
      {	// use tasking for children and for extend-add parallelism
	multifrontal_factorization_phase1(strat, etree_level, task_depth, true);
      }
      // do not use tasking for blas/lapack parallelism (use system blas threading!)
      multifrontal_factorization_phase2(strat, etree_level, task_depth, false);
    } else {
      bool tasked = task_depth < params::task_recursion_cutoff_level;
      multifrontal_factorization_phase1(strat, etree_level, task_depth, tasked);
      multifrontal_factorization_phase2(strat, etree_level, task_depth, tasked);
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDense<scalar_t,real_t,integer_t>::multifrontal_factorization_phase1
  (HSSStrategy<real_t>& strat, int etree_level, int task_depth, bool tasked) {
    if (tasked) {
      if (this->lchild)
#pragma omp task untied default(shared) final(task_depth >= params::task_recursion_cutoff_level-1) mergeable
	this->lchild->multifrontal_factorization(strat, etree_level+1, task_depth+1);
      if (this->rchild)
#pragma omp task untied default(shared) final(task_depth >= params::task_recursion_cutoff_level-1) mergeable
	this->rchild->multifrontal_factorization(strat, etree_level+1, task_depth+1);
#pragma omp taskwait
    } else {
      if (this->lchild) this->lchild->multifrontal_factorization(strat, etree_level+1, task_depth);
      if (this->rchild) this->rchild->multifrontal_factorization(strat, etree_level+1, task_depth);
    }
    F11 = new scalar_t[this->dim_sep*this->dim_sep+2*this->dim_sep*this->dim_upd];
    F12 = F11+this->dim_sep*this->dim_sep;
    F21 = F12+this->dim_upd*this->dim_sep;
    piv = new int[this->dim_sep];
    std::fill(F11, F11+this->dim_sep*this->dim_sep+2*this->dim_sep*this->dim_upd, scalar_t(0.));
    std::fill(piv, piv+this->dim_sep, 0);
    this->A->extract_front(F11, F12, F21, this->dim_sep, this->dim_upd,
			   this->sep_begin, this->sep_end, this->upd, task_depth);
    if (this->dim_upd) {
      F22 = new scalar_t[this->dim_upd*this->dim_upd];
      std::fill(F22, F22+this->dim_upd*this->dim_upd, scalar_t(0.));
    }
    if (this->lchild) {
      extend_add(static_cast<FrontalMatrixDense<scalar_t,real_t,integer_t>* >(this->lchild), task_depth);
      this->lchild->release_work_memory();
    }
    if (this->rchild) {
      extend_add(static_cast<FrontalMatrixDense<scalar_t,real_t,integer_t>* >(this->rchild), task_depth);
      this->rchild->release_work_memory();
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDense<scalar_t,real_t,integer_t>::multifrontal_factorization_phase2
  (HSSStrategy<real_t>& strat, int etree_level, int task_depth, bool tasked) {
    if (this->dim_sep) {
      int info;    // partial LU decomposition of F11 block
      if (tasked) {
	getrf_omp_task(this->dim_sep, this->dim_sep, F11, this->dim_sep, piv, &info, task_depth);
	if (info) {
	  std::cerr << "In FrontalMatrixDense::multifrontal_factorization:"
		    << " LU factorization failed with info = " << info << std::endl;
	  std::cerr << "This is probably due to a zero pivot!" << std::endl
		    << "Try running with MC64 pivoting (--sp_mc64job 5)" << std::endl;
	  exit(1);
	}
	if (this->dim_upd) {
#pragma omp task
	  { laswp_omp_task(this->dim_upd, F12, this->dim_sep, 1, this->dim_sep, piv, 1, task_depth);
	    trsm_omp_task('L', 'L', 'N', 'U', this->dim_sep, this->dim_upd, scalar_t(1.), F11, this->dim_sep, F12, this->dim_sep, task_depth); }
	  trsm_omp_task('R', 'U', 'N', 'N', this->dim_upd, this->dim_sep, scalar_t(1.), F11, this->dim_sep, F21, this->dim_upd, task_depth);
#pragma omp taskwait
	  gemm_omp_task('N', 'N', this->dim_upd, this->dim_upd, this->dim_sep, scalar_t(-1.), F21, this->dim_upd,
			F12, this->dim_sep, scalar_t(1.), F22, this->dim_upd, task_depth);
	}
      } else {
	blas::getrf(this->dim_sep, this->dim_sep, F11, this->dim_sep, piv, &info);
	if (info) {
	  std::cerr << "In FrontalMatrixDense::multifrontal_factorization:"
		    << " LU factorization failed with info = " << info << std::endl;
	  std::cerr << "This is probably due to a zero pivot!" << std::endl
		    << "Try running with MC64 pivoting (--sp_mc64job 5)" << std::endl;
	  exit(1);
	}
	if (this->dim_upd) {
	  blas::laswp(this->dim_upd, F12, this->dim_sep, 1, this->dim_sep, piv, 1);
	  blas::trsm('L', 'L', 'N', 'U', this->dim_sep, this->dim_upd, scalar_t(1.), F11, this->dim_sep, F12, this->dim_sep);
	  blas::trsm('R', 'U', 'N', 'N', this->dim_upd, this->dim_sep, scalar_t(1.), F11, this->dim_sep, F21, this->dim_upd);
	  blas::gemm('N', 'N', this->dim_upd, this->dim_upd, this->dim_sep, scalar_t(-1.), F21, this->dim_upd,
		     F12, this->dim_sep, scalar_t(1.), F22, this->dim_upd);
	}
      }
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDense<scalar_t,real_t,integer_t>::forward_multifrontal_solve
  (scalar_t* b, scalar_t* wmem, int etree_level, int task_depth) {
    if (task_depth == 0) {
#pragma omp parallel if(!omp_in_parallel())
#pragma omp single
      { // tasking when calling the children
 	forward_multifrontal_solve_phase1(b, wmem, etree_level, task_depth, true);
      }
      // no tasking for the root node computations, use system blas threading!
       forward_multifrontal_solve_phase2(b, wmem, etree_level, task_depth, false);
    } else {
      bool tasked = task_depth < params::task_recursion_cutoff_level;
      forward_multifrontal_solve_phase1(b, wmem, etree_level, task_depth, tasked);
      forward_multifrontal_solve_phase2(b, wmem, etree_level, task_depth, tasked);
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDense<scalar_t,real_t,integer_t>::forward_multifrontal_solve_phase1
  (scalar_t* b, scalar_t* wmem, int etree_level, int task_depth, bool tasked) {
    if (tasked) {
      if (this->lchild)
#pragma omp task untied default(shared) final(task_depth >= params::task_recursion_cutoff_level-1) mergeable
	this->lchild->forward_multifrontal_solve(b, wmem, etree_level+1, task_depth+1);
      if (this->rchild)
#pragma omp task untied default(shared) final(task_depth >= params::task_recursion_cutoff_level-1) mergeable
	this->rchild->forward_multifrontal_solve(b, wmem, etree_level+1, task_depth+1);
#pragma omp taskwait
    } else {
      if (this->lchild) this->lchild->forward_multifrontal_solve(b, wmem, etree_level+1, task_depth);
      if (this->rchild) this->rchild->forward_multifrontal_solve(b, wmem, etree_level+1, task_depth);
    }
    this->look_left(b, wmem);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDense<scalar_t,real_t,integer_t>::forward_multifrontal_solve_phase2
  (scalar_t* b, scalar_t* wmem, int etree_level, int task_depth, bool tasked) {
    if (this->dim_sep) {
      blas::laswp(1, b+this->sep_begin, this->A->size(), 1, this->dim_sep, piv, 1);
      if (tasked) trsv_omp_task('L', 'N', 'U', this->dim_sep, F11, this->dim_sep, b+this->sep_begin, 1, task_depth);
      else blas::trsv('L', 'N', 'U', this->dim_sep, F11, this->dim_sep, b+this->sep_begin, 1);
      if (this->dim_upd) {
	if (tasked) gemv_omp_task('N', this->dim_upd, this->dim_sep, scalar_t(-1.), F21, this->dim_upd,
				  b+this->sep_begin, 1, scalar_t(1.), wmem+this->p_wmem, 1, task_depth);
	else blas::gemv('N', this->dim_upd, this->dim_sep, scalar_t(-1.), F21, this->dim_upd,
			b+this->sep_begin, 1, scalar_t(1.), wmem+this->p_wmem, 1);
      }
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDense<scalar_t,real_t,integer_t>::backward_multifrontal_solve
  (scalar_t* b, scalar_t* wmem, int etree_level, int task_depth) {
    if (task_depth == 0) {
      // no tasking in blas routines, use system threaded blas instead
      backward_multifrontal_solve_phase1(b, wmem, etree_level, task_depth, false);
#pragma omp parallel if(!omp_in_parallel())
#pragma omp single
      // tasking when calling children
      backward_multifrontal_solve_phase2(b, wmem, etree_level, task_depth, true);
    } else {
      bool tasked = task_depth < params::task_recursion_cutoff_level;
      backward_multifrontal_solve_phase1(b, wmem, etree_level, task_depth, tasked);
      backward_multifrontal_solve_phase2(b, wmem, etree_level, task_depth, tasked);
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDense<scalar_t,real_t,integer_t>::backward_multifrontal_solve_phase1
  (scalar_t* y, scalar_t* wmem, int etree_level, int task_depth, bool tasked) {
    if (this->dim_sep) {
      if (this->dim_upd > 0) {
	if (tasked) gemv_omp_task('N', this->dim_sep, this->dim_upd, scalar_t(-1.), F12, this->dim_sep,
				  wmem+this->p_wmem, 1, scalar_t(1.), y+this->sep_begin, 1, task_depth);
	else blas::gemv('N', this->dim_sep, this->dim_upd, scalar_t(-1.), F12, this->dim_sep,
			wmem+this->p_wmem, 1, scalar_t(1.), y+this->sep_begin, 1);
      }
      if (tasked) trsv_omp_task('U', 'N', 'N', this->dim_sep, F11, this->dim_sep, y+this->sep_begin, 1, task_depth);
      else blas::trsv('U', 'N', 'N', this->dim_sep, F11, this->dim_sep, y+this->sep_begin, 1);
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDense<scalar_t,real_t,integer_t>::backward_multifrontal_solve_phase2
  (scalar_t* y, scalar_t* wmem, int etree_level, int task_depth, bool tasked) {
    this->look_right(y, wmem);
    if (tasked) {
      if (this->lchild)
#pragma omp task untied default(shared) final(task_depth >= params::task_recursion_cutoff_level-1) mergeable
	this->lchild->backward_multifrontal_solve(y, wmem, etree_level+1, task_depth+1);
      if (this->rchild)
#pragma omp task untied default(shared) final(task_depth >= params::task_recursion_cutoff_level-1) mergeable
	this->rchild->backward_multifrontal_solve(y, wmem, etree_level+1, task_depth+1);
#pragma omp taskwait
    } else {
      if (this->lchild) this->lchild->backward_multifrontal_solve(y, wmem, etree_level+1, task_depth);
      if (this->rchild) this->rchild->backward_multifrontal_solve(y, wmem, etree_level+1, task_depth);
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrixDense<scalar_t,real_t,integer_t>::get_submatrix
  (integer_t* rows, integer_t nr_rows, integer_t* cols, integer_t nr_cols,
   scalar_t* values, integer_t ld, int task_depth) {
    std::vector<std::pair<integer_t,integer_t>> myrows; myrows.reserve(nr_rows);
    for (integer_t r=0; r<nr_rows; r++) {
      integer_t* l = std::lower_bound(this->upd, this->upd+this->dim_upd, rows[r]);
      if (l != this->upd+this->dim_upd && *l == rows[r])
	myrows.push_back(std::make_pair(l-this->upd,r));
    }
    if (myrows.empty()) return;
    long long int local_flops = 0;
    for (integer_t c=0; c<nr_cols; c++) {
      integer_t* l = std::lower_bound(this->upd, this->upd+this->dim_upd, cols[c]);
      if (l != this->upd+this->dim_upd && *l == cols[c]) {
	for (auto r : myrows)
	  values[std::get<1>(r)+c*ld] += F22[std::get<0>(r)+(l-this->upd)*this->dim_upd];
	local_flops += myrows.size();
      }
    }
    STRUMPACK_FLOPS((is_complex<scalar_t>() ? 2 : 1) * local_flops);
  }

} // end namespace strumpack

#endif
