/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */
#ifndef HSS_MATRIX_HPP
#define HSS_MATRIX_HPP

#include <iostream>
#include <iomanip>
#include <cassert>
#include "tools.hpp"
#include "ID.hpp"
#include "blas_lapack_wrapper.hpp"
#include "TaskTimer.hpp"
#include "strumpack_parameters.hpp"
#include "MatrixReordering.hpp"

namespace strumpack {

  template<typename scalar_t,typename integer_t> class WorkmemTree {
  public:
    std::vector<WorkmemTree*> c;
    scalar_t* wmemu;
    scalar_t* wmemv;
    integer_t flag;

    WorkmemTree() : wmemu(NULL), wmemv(NULL), flag(0) {}
    ~WorkmemTree() {
      for (auto ch : c) delete ch;
      delete[] wmemu;
    }
    void allocate_children() {
      c.push_back(new WorkmemTree());
      c.push_back(new WorkmemTree());
    }
    void allocate_wmem(integer_t su, integer_t sv) {
      wmemu = new scalar_t[su+sv];
      wmemv = wmemu+su;
    }
  };

  template<typename scalar_t,typename real_t,typename integer_t> class HSSMatrix {
  public:
    integer_t u_rows;
    integer_t v_rows;
    integer_t u_rank;
    integer_t v_rank;

    integer_t subset_begin;
    integer_t subset_end;

    std::vector<HSSMatrix<scalar_t,real_t,integer_t>* > child;

    scalar_t* D;
    scalar_t* B;
    scalar_t* L;      // lower triangular matrix from ULV factorization, W_{\tau;2} is stored below L
    scalar_t* Q;

    scalar_t* U;      // data for the U row basis
    scalar_t* V;      // data for the V column basis
    scalar_t* Vt;     // data for the reduced V column basis

    int* U_perm;
    int* V_perm;

    int* locr;      // used in compression, in ULV for the root this contains
    // the permutation of the PLU decomposition of \tilde{D}
    int* locc;

    scalar_t* work_u;
    scalar_t* work_v;
    integer_t work_u_nnz;
    integer_t work_v_nnz;

    int flag;        // can be used for different purposes
    enum state_name {UNTOUCHED, PARTIALLY_COMPRESSED, COMPRESSED};
    state_name state;

    HSSMatrix();
    HSSMatrix(integer_t dim);
    HSSMatrix(HSSPartitionTree& part_tree, integer_t dim_right, integer_t set_begin);
    HSSMatrix(HSSPartitionTree& part_tree, integer_t set_begin);
    ~HSSMatrix();

    template<typename Mat>
    void compress(Mat* F, HSSStrategy<real_t>& strat, integer_t random_cols, scalar_t* Rrow, scalar_t* Rcol, integer_t ldr,
		  scalar_t* Scol, scalar_t* Srow, integer_t lds, int depth=0, bool isroot=true);
    void ULV_factorization(int depth, bool isroot);
    void ULV_forward_solve(scalar_t* f, scalar_t* reduced_rhs, int depth, bool isroot);
    void ULV_backward_solve(scalar_t* f, int depth, bool isroot);
    void matvec(scalar_t* x, integer_t ldx, scalar_t beta, scalar_t* y, integer_t ldy, integer_t nrhs, int depth=0);
    void matvec_transpose(scalar_t* x, integer_t ldx, scalar_t beta, scalar_t* y, integer_t ldy, integer_t nrhs, int depth=0);
    scalar_t get_ij(integer_t row, integer_t col);
    void extract_submatrix(std::vector<integer_t>& rows, std::vector<integer_t>& cols,
			   std::vector<integer_t>& orig_rows, std::vector<integer_t>& orig_cols,
			   scalar_t* values, integer_t ld, integer_t dim_upd, integer_t dim_sep, int depth);
    void fast_Schur_update(scalar_t* theta, integer_t ldt, scalar_t* phi, integer_t ldp, int depth);
    void destroy_update_block();
    integer_t get_maximum_rank();
    long long get_factor_nonzeros();
    void print();
    void print_ranks(std::ostream &out, bool isroot=true);

  private:
    HSSMatrix(integer_t set_begin, integer_t set_size);
    HSSMatrix(const HSSMatrix&) = delete;
    HSSMatrix& operator=(HSSMatrix const&) = delete;

    void release_workmemory();
    void reallocate_workmemory(integer_t work_u_size, integer_t work_v_size);
    void matvec_forward_loop(scalar_t* x, integer_t ldx, integer_t nrhs, int depth);
    void matvec_backward_loop(scalar_t* x, integer_t ldx, scalar_t beta, scalar_t* y, integer_t ldy, integer_t nrhs, bool isroot, int depth);
    void matvec_transpose_forward_loop(scalar_t* x, integer_t ldx, integer_t nrhs, int depth);
    void matvec_transpose_backward_loop(scalar_t* x, integer_t ldx, scalar_t beta, scalar_t* y, integer_t ldy, integer_t nrhs, bool isroot, int depth);
    void apply_U(bool transpose, integer_t n, scalar_t* x, integer_t ldx, scalar_t beta, scalar_t* y, integer_t ldy, int depth);
    void apply_V(bool transpose, integer_t n, scalar_t* x, integer_t ldx, scalar_t beta, scalar_t* y, integer_t ldy, int depth);
    void apply_U(bool transpose, scalar_t* x, scalar_t beta, scalar_t* y, int depth);
    void apply_V(bool transpose, scalar_t* x, scalar_t beta, scalar_t* y, int depth);
    void apply_U_permutation_forward(integer_t cols, scalar_t* x, integer_t ldx);
    void apply_U_permutation_backward(integer_t cols, scalar_t* x, integer_t ldx);
    void apply_V_permutation_forward(integer_t cols, scalar_t* x, integer_t ldx);
    void apply_V_permutation_backward(integer_t cols, scalar_t* x, integer_t ldx);
    void apply_U_permutation_forward(scalar_t* x);
    void apply_U_permutation_backward(scalar_t* x);
    void apply_V_permutation_forward(scalar_t* x);
    void apply_V_permutation_backward(scalar_t* x);
    void extract_forward_loop(std::vector<integer_t>& my_cols, WorkmemTree<scalar_t,integer_t>* w, int depth);
    void extract_backward_loop(std::vector<integer_t>& my_rows, std::vector<integer_t>& my_cols,
			       std::vector<integer_t>& orig_rows, std::vector<integer_t>& orig_cols,
			       scalar_t* values, integer_t ld, WorkmemTree<scalar_t,integer_t>* w,
			       bool isroot, int depth);
    void apply_UV_big(integer_t Uop_cols, scalar_t* Uop, integer_t ldu,
		      integer_t Vop_cols, scalar_t* Vop, integer_t ldv, scalar_t beta, int depth);
  };

} // end namespace strumpack

#include "HSSMatrix_matvec.hpp"
#include "HSSMatrix_ULV.hpp"
#include "HSSMatrix_compress.hpp"
#include "HSSMatrix_extract.hpp"

namespace strumpack {

  template<typename scalar_t,typename real_t,typename integer_t>
  HSSMatrix<scalar_t,real_t,integer_t>::HSSMatrix()
    : u_rows(0), v_rows(0), u_rank(0), v_rank(0), subset_begin(0), subset_end(0),
      D(NULL), B(NULL), L(NULL), Q(NULL), U(NULL), V(NULL), Vt(NULL),
      U_perm(NULL), V_perm(NULL), locr(NULL), locc(NULL), work_u(NULL), work_v(NULL),
      work_u_nnz(0), work_v_nnz(0), state(UNTOUCHED) {
  }

  // This is used to create the top node of an HSS tree split in a separator and update part
  template<typename scalar_t,typename real_t,typename integer_t>
  HSSMatrix<scalar_t,real_t,integer_t>::HSSMatrix
  (HSSPartitionTree& part_tree, integer_t dim_right, integer_t set_begin) : HSSMatrix() {
    child.push_back(new HSSMatrix<scalar_t,real_t,integer_t>(part_tree, set_begin));
    child.push_back(new HSSMatrix<scalar_t,real_t,integer_t>(set_begin+part_tree.size, dim_right));
    subset_begin = child[0]->subset_begin;
    subset_end = child[1]->subset_end;
  }

  template<typename scalar_t,typename real_t,typename integer_t>
  HSSMatrix<scalar_t,real_t,integer_t>::HSSMatrix(HSSPartitionTree& part_tree, integer_t set_begin) : HSSMatrix() {
    subset_begin = set_begin;
    subset_end = set_begin + part_tree.size;
    for (auto ch : part_tree.c) {
      child.push_back(new HSSMatrix<scalar_t,real_t,integer_t>(ch, set_begin));
      set_begin += ch.size;
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t>
  HSSMatrix<scalar_t,real_t,integer_t>::HSSMatrix(integer_t dim) : HSSMatrix(0, dim) {}

  // create an HSS matrix of size set_size by recursively splitting
  // while set_size > strumpack::hss_minimum_partition_size
  template<typename scalar_t,typename real_t,typename integer_t>
  HSSMatrix<scalar_t,real_t,integer_t>::HSSMatrix(integer_t set_begin, integer_t set_size) : HSSMatrix() {
    integer_t cf = params::hss_coarsening_factor;
    if (set_size < params::hss_minimum_partition_size || set_size == 1) {
      subset_begin = set_begin;
      subset_end = set_begin + set_size;
    } else {
      for (integer_t i=0; i<cf-1; i++)
	child.push_back(new HSSMatrix<scalar_t,real_t,integer_t>(set_begin + i*set_size/cf, set_size/cf));
      child.push_back(new HSSMatrix<scalar_t,real_t,integer_t>(set_begin + (cf-1)*set_size/cf, set_size - (cf-1)*set_size/cf));
      subset_begin = child[0]->subset_begin;
      subset_end = child[cf-1]->subset_end;
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t>
  HSSMatrix<scalar_t,real_t,integer_t>::~HSSMatrix() {
    delete[] D;    delete[] B;      delete[] U;      delete[] V;      delete[] Vt;
    delete[] L;    delete[] Q;      delete[] locr;
    delete[] work_u; delete[] work_v;
    for (auto ch : child) delete ch;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::destroy_update_block() {
    if (!child.empty()) {
      delete child.back();
      child.pop_back();    // delete the contribution block (the update part)
    }
    delete[] B;  B = NULL;
    release_workmemory();
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::release_workmemory() {
    delete[] work_u; work_u = NULL; work_u_nnz = 0;
    delete[] work_v; work_v = NULL; work_v_nnz = 0;
    for (auto ch : child) ch->release_workmemory();
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::reallocate_workmemory(integer_t work_u_size, integer_t work_v_size) {
    if (work_u_size > work_u_nnz) {
      work_u_nnz = work_u_size;
      delete[] work_u; work_u = new scalar_t[work_u_nnz];
    }
    if (work_v_size > work_v_nnz) {
      work_v_nnz = work_v_size;
      delete[] work_v; work_v = new scalar_t[work_v_nnz];
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::apply_U_permutation_forward(integer_t cols, scalar_t* x, integer_t ldx) {
    for (integer_t i=u_rows-1; i>=0; i--) if (U_perm[i]!=i) blas::swap(cols, &x[i], ldx, &x[U_perm[i]], ldx);
  }
  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::apply_U_permutation_backward(integer_t cols, scalar_t* x, integer_t ldx) {
    for (integer_t i=0; i<u_rows; i++) if (U_perm[i]!=i) blas::swap(cols, &x[i], ldx, &x[U_perm[i]], ldx);
  }
  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::apply_V_permutation_forward(integer_t cols, scalar_t* x, integer_t ldx) {
    for (integer_t i=v_rows-1; i>=0; i--) if (V_perm[i]!=i) blas::swap(cols, &x[i], ldx, &x[V_perm[i]], ldx);
  }
  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::apply_V_permutation_backward(integer_t cols, scalar_t* x, integer_t ldx) {
    for (integer_t i=0; i<v_rows; i++) if (V_perm[i]!=i) blas::swap(cols, &x[i], ldx, &x[V_perm[i]], ldx);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::apply_U_permutation_forward(scalar_t* x) {
    for (integer_t i=u_rows-1; i>=0; i--) if (U_perm[i]!=i) std::swap(x[i], x[U_perm[i]]);
  }
  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::apply_U_permutation_backward(scalar_t* x) {
    for (integer_t i=0; i<u_rows; i++) if (U_perm[i]!=i) std::swap(x[i], x[U_perm[i]]);
  }
  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::apply_V_permutation_forward(scalar_t* x) {
    for (integer_t i=v_rows-1; i>=0; i--) if (V_perm[i]!=i) std::swap(x[i], x[V_perm[i]]);
  }
  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::apply_V_permutation_backward(scalar_t* x) {
    for (integer_t i=0; i<v_rows; i++) if (V_perm[i]!=i) std::swap(x[i], x[V_perm[i]]);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::apply_U
  (bool transpose, integer_t n, scalar_t* x, integer_t ldx, scalar_t beta, scalar_t* y, integer_t ldy, int depth) {
    if (u_rank) {
      bool tasked = depth<params::task_recursion_cutoff_level;
      if (transpose) {
	apply_U_permutation_backward(n, x, ldx);
	if (tasked) gemm_omp_task('C', 'N', u_rank, n, u_rows-u_rank, scalar_t(1.), &U[u_rank], u_rows, &x[u_rank], ldx, beta, y, ldy, depth);
	else blas::gemm('C', 'N', u_rank, n, u_rows-u_rank, scalar_t(1.), &U[u_rank], u_rows, &x[u_rank], ldx, beta, y, ldy);
	mat_add(u_rank, n, x, ldx, scalar_t(1.), y, ldy);
	apply_U_permutation_forward(n, x, ldx);
      } else {
	if (beta != scalar_t(0.)) apply_U_permutation_backward(n, y, ldy);
	if (tasked) gemm_omp_task('N', 'N', u_rows-u_rank, n, u_rank, scalar_t(1.), &U[u_rank], u_rows, x, ldx, beta, &y[u_rank], ldy, depth);
	else blas::gemm('N', 'N', u_rows-u_rank, n, u_rank, scalar_t(1.), &U[u_rank], u_rows, x, ldx, beta, &y[u_rank], ldy);
	if (beta == scalar_t(0.)) blas::lacpy('A', u_rank, n, x, ldx, y, ldy);
	else mat_add(u_rank, n, x, ldx, beta, y, ldy);
	apply_U_permutation_forward(n, y, ldy);
      }
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::apply_V
  (bool transpose, integer_t n, scalar_t* x, integer_t ldx, scalar_t beta, scalar_t* y, integer_t ldy, int depth) {
    if (v_rank) {
      bool tasked = depth<params::task_recursion_cutoff_level;
      if (transpose) {
	apply_V_permutation_backward(n, x, ldx);
	if (tasked) gemm_omp_task('C', 'N', v_rank, n, v_rows-v_rank, scalar_t(1.), &V[v_rank], v_rows, &x[v_rank], ldx, beta, y, ldy, depth);
	else blas::gemm('C', 'N', v_rank, n, v_rows-v_rank, scalar_t(1.), &V[v_rank], v_rows, &x[v_rank], ldx, beta, y, ldy);
	mat_add(v_rank, n, x, ldx, scalar_t(1.), y, ldy);
	apply_V_permutation_forward(n, x, ldx);
      } else {
	if (beta != scalar_t(0.)) apply_V_permutation_backward(n, y, ldy);
	if (tasked) gemm_omp_task('N', 'N', v_rows-v_rank, n, v_rank, scalar_t(1.), &V[v_rank], v_rows, x, ldx, beta, &y[v_rank], ldy, depth);
	else blas::gemm('N', 'N', v_rows-v_rank, n, v_rank, scalar_t(1.), &V[v_rank], v_rows, x, ldx, beta, &y[v_rank], ldy);
	if (beta == scalar_t(0.)) blas::lacpy('A', v_rank, n, x, ldx, y, ldy);
	else mat_add(v_rank, n, x, ldx, beta, y, ldy);
	apply_V_permutation_forward(n, y, ldy);
      }
    }
  }


  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::apply_U
  (bool transpose, scalar_t* x, scalar_t beta, scalar_t* y, int depth) {
    if (u_rank) {
      if (transpose) {
	apply_U_permutation_backward(x);
	gemv_omp_task('C', u_rows-u_rank, u_rank, scalar_t(1.), U+u_rank, u_rows, x+u_rank, 1, beta, y, 1, depth);
	mat_add(u_rank, 1, x, 1, scalar_t(1.), y, 1);
	apply_U_permutation_forward(x);
      } else {
	if (beta != scalar_t(0.)) apply_U_permutation_backward(y);
	gemv_omp_task('N', u_rows-u_rank, u_rank, scalar_t(1.), U+u_rank, u_rows, x, 1, beta, y+u_rank, 1, depth);
	if (beta == scalar_t(0.)) std::copy(x, x+u_rank, y);
	else mat_add(u_rank, 1, x, 1, beta, y, 1);
	apply_U_permutation_forward(y);
      }
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::apply_V
  (bool transpose, scalar_t* x, scalar_t beta, scalar_t* y, int depth) {
    if (v_rank) {
      if (transpose) {
	apply_V_permutation_backward(x);
	gemv_omp_task('C', v_rows-v_rank, v_rank, scalar_t(1.), V+v_rank, v_rows, x+v_rank, 1, beta, y, 1, depth);
	mat_add<scalar_t,integer_t>(v_rank, 1, x, 1, scalar_t(1.), y, 1);
	apply_V_permutation_forward(x);
      } else {
	if (beta != scalar_t(0.)) apply_V_permutation_backward(y);
	gemv_omp_task('N', v_rows-v_rank, v_rank, scalar_t(1.), V+v_rank, v_rows, x, 1, beta, y+v_rank, 1, depth);
	if (beta == scalar_t(0.)) std::copy(x, x+v_rank, y);
	else mat_add<scalar_t,integer_t>(v_rank, 1, x, 1, beta, y, 1);
	apply_V_permutation_forward(y);
      }
    }
  }

  // workmem should be same size as theta/phi ?
  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::fast_Schur_update
  (scalar_t* theta, integer_t ldt, scalar_t* phi, integer_t ldp, int depth) {
    HSSMatrix<scalar_t,real_t,integer_t> *c0 = child[0], *c1 = child[1];
    if (c0->u_rank && c0->v_rank && c1->u_rank && c1->v_rank) {
      bool tasked = depth<params::task_recursion_cutoff_level;
      scalar_t* tmp = new scalar_t[c0->u_rows*c1->v_rank];
      c1->reallocate_workmemory(c1->u_rank*c0->u_rows, c1->v_rank*c0->u_rows);
      c0->apply_U(false, c1->v_rank, &B[c1->u_rank*c0->v_rank], c0->u_rank, scalar_t(0.), tmp, c0->u_rows, depth);
      if (tasked)
	getrs_omp_task('N', c0->u_rows, c1->v_rank, c0->D, c0->u_rows, c0->locr, tmp, c0->u_rows, &flag, depth);
      else
	blas::getrs('N', c0->u_rows, c1->v_rank, c0->D, c0->u_rows, c0->locr, tmp, c0->u_rows, &flag);
      blas::omatcopy('C', c0->u_rows, c1->v_rank, tmp, c0->u_rows, c1->work_v, c1->v_rank);
      blas::omatcopy('N', c1->u_rank, c0->v_rank, B, c1->u_rank, c1->work_u, c1->u_rank);
      delete[] tmp;
      c1->apply_UV_big(c0->v_rank, theta, ldt, c0->u_rows, phi, ldp, scalar_t(0.), depth);
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::apply_UV_big
  (integer_t Uop_cols, scalar_t* Uop, integer_t ldu,
   integer_t Vop_cols, scalar_t* Vop, integer_t ldv, scalar_t beta, int depth) {
    bool tasked = depth<params::task_recursion_cutoff_level;
    if (child.empty()) {
      if (tasked) {
#pragma omp task
	apply_U(false, Uop_cols, work_u, u_rank, beta, Uop+subset_begin, ldu, depth); // theta = U * theta
	apply_V(false, Vop_cols, work_v, v_rank, beta, Vop+subset_begin, ldv, depth);
#pragma omp taskwait
      } else {
	apply_U(false, Uop_cols, work_u, u_rank, beta, Uop+subset_begin, ldu, depth);
	apply_V(false, Vop_cols, work_v, v_rank, beta, Vop+subset_begin, ldv, depth);
      }
    } else {
      for (auto ch : child) ch->reallocate_workmemory(ch->u_rank*Uop_cols, ch->v_rank*Vop_cols);
      scalar_t* tmp = new scalar_t[std::max(u_rows*Uop_cols,v_rows*Vop_cols)];
      if (u_rank) {
	apply_U(false, Uop_cols, work_u, u_rank, scalar_t(0.), tmp, u_rows, depth);
	integer_t pos = 0;
	for (auto ch : child) {
	  blas::lacpy('A', ch->u_rank, Uop_cols, &tmp[pos], u_rows, ch->work_u, ch->u_rank);
	  pos += ch->u_rank;
	}
      }
      if (v_rank) {
	apply_V(false, Vop_cols, work_v, v_rank, scalar_t(0.), tmp, v_rows, depth);
	integer_t pos = 0;
	for (auto ch : child) {
	  blas::lacpy('A', ch->v_rank, Vop_cols, tmp+pos, v_rows, ch->work_v, ch->v_rank);
	  pos += ch->v_rank;
	}
      }
      delete[] tmp;
      if (tasked) {
	for (auto ch : child)
#pragma omp task firstprivate(ch) default(shared)
	  ch->apply_UV_big(Uop_cols, Uop, ldu, Vop_cols, Vop, ldv, beta, depth+1);
#pragma omp taskwait
      } else {
	for (auto ch : child) ch->apply_UV_big(Uop_cols, Uop, ldu, Vop_cols, Vop, ldv, beta, depth);
      }
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> integer_t
  HSSMatrix<scalar_t,real_t,integer_t>::get_maximum_rank() {
    integer_t max_rank = std::max(u_rank, v_rank);
    for (auto ch : child) if(ch) max_rank = std::max(ch->get_maximum_rank(), max_rank);
    return max_rank;
  }

  template<typename scalar_t,typename real_t,typename integer_t> long long
  HSSMatrix<scalar_t,real_t,integer_t>::get_factor_nonzeros() {
    long long mem = 0.;
    if (D) mem += u_rows*u_rows;
    if (U && U_perm) mem += u_rows + (u_rows-u_rank)*u_rank;
    if (V && V_perm) mem += v_rows + (v_rows-v_rank)*v_rank;
    if (locr) mem += u_rows;
    if (locc) mem += v_rows;
    if (B) for (auto ci : child) for (auto cj : child) if (ci != cj) mem += ci->u_rank*cj->v_rank;
    if (L) mem += u_rows*u_rows;
    if (Q) mem += u_rows*u_rows;
    if (Vt) mem += u_rows*v_rank;
    for (auto ch : child) if (ch) mem += ch->get_factor_nonzeros();
    return mem;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::print() {
    auto old_precision = std::cout.precision(15);
    std::cout << " = [ ";
    for (integer_t r=subset_begin; r<subset_end; r++) {
      for (integer_t c=subset_begin; c<subset_end; c++) {
	scalar_t v = get_ij(r, c);
	if (std::abs(v) < 1e-10) v = scalar_t(0.);
	std::cout << v << " ";
      } std::cout << ";\n";
    } std::cout << " ];" << std::endl;
    std::cout.precision(old_precision);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::print_ranks(std::ostream &out, bool isroot) {
    for (auto ch : child) if (ch) ch->print_ranks(out, false);
    if (!isroot) {
      out << "b = [" << std::setw(5) << subset_begin << "," << subset_end << "]"
	  << "  U = " << std::setw(5) << u_rows << " x " << std::setw(5) << u_rank
	  << "  V = " << std::setw(5) << v_rows << " x " << std::setw(5) << v_rank << "  ";
      if (child.empty()) out << "leaf" << std::endl;
      else out << "non-leaf" << std::endl;
    }
  }

} // end namespace strumpack

#endif // HSS_MATRIX_HPP
