/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */
#ifndef HSS_MATRIX_COMPRESS_HPP
#define HSS_MATRIX_COMPRESS_HPP

#include "ID.hpp"

namespace strumpack {

  /**
   * This implements
   * "Algorithm 4. Computing the HSS factorization of a nonsymmetric matrix."
   * from
   * P.G. Martinsson, "A fast randomized algorithm for computing
   *  a hierarchically semiseparable representation of a matrix",
   *  SIAM, J. Matrix Anal. Appl., Vol.32, No. 4, pp. 1251-1274
   */
  template<typename scalar_t,typename real_t,typename integer_t> template<typename Mat> void
  HSSMatrix<scalar_t,real_t,integer_t>::compress
  (Mat* F, HSSStrategy<real_t>& strat, integer_t random_cols, scalar_t* Rrow, scalar_t* Rcol, integer_t ldr,
   scalar_t* Scol, scalar_t* Srow, integer_t lds, int depth, bool isroot) {
    bool tasked = depth<params::task_recursion_cutoff_level;
    scalar_t* Srow_loc; scalar_t* Scol_loc; scalar_t* Rrow_loc; scalar_t* Rcol_loc;
    integer_t random_cols_loc, old_random_cols;
    if (state==COMPRESSED || state==PARTIALLY_COMPRESSED) {
      // if this node is compressed or compression was started but failed
      // then only update the new random vectors
      old_random_cols = random_cols / 2;
      random_cols_loc = random_cols / 2;
      Srow_loc = Srow + subset_begin + old_random_cols*lds; Scol_loc = Scol + subset_begin + old_random_cols*lds;
      Rrow_loc = Rrow + subset_begin + old_random_cols*ldr; Rcol_loc = Rcol + subset_begin + old_random_cols*ldr;
    } else {
      old_random_cols = 0;
      random_cols_loc = random_cols;
      Srow_loc = Srow + subset_begin; Scol_loc = Scol + subset_begin;
      Rrow_loc = Rrow + subset_begin; Rcol_loc = Rcol + subset_begin;
    }
    integer_t* indc = F->col_ind;
    integer_t* indr = F->row_ind;

    if (child.empty()) { // leaf
      u_rows = subset_end - subset_begin;
      v_rows = u_rows;
      if (D==NULL) {
	D = new scalar_t[u_rows*u_rows]; 	// save diagonal block, extract A(I_t, I_t)
	F->extract(&indr[subset_begin], u_rows, &indc[subset_begin], v_rows, D, u_rows, depth);
      }
      if (u_rows && v_rows && !isroot) {
	if (tasked) {
#pragma omp task default(shared)
	  gemm_omp_task('C', 'N', v_rows, random_cols_loc, u_rows, scalar_t(-1.0), D, u_rows,  // Scol = Scol(I_t,:) - A(I_t,I_t)^T Rrow
			Rcol_loc, ldr, scalar_t(1.0), Scol_loc, lds, depth);
	  gemm_omp_task('N', 'N', u_rows, random_cols_loc, v_rows, scalar_t(-1.0), D, u_rows,  // Srow = Srow(I_t,:) - A(I_t,I_t) Rrow
			Rrow_loc, ldr, scalar_t(1.0), Srow_loc, lds, depth);
#pragma omp taskwait
	} else {
	  blas::gemm('N', 'N', u_rows, random_cols_loc, v_rows, scalar_t(-1.0), D, u_rows,
		     Rrow_loc, ldr, scalar_t(1.0), Srow_loc, lds);
	  blas::gemm('C', 'N', v_rows, random_cols_loc, u_rows, scalar_t(-1.0), D, u_rows,
		     Rcol_loc, ldr, scalar_t(1.0), Scol_loc, lds);
	}
      }
    } else { // non-leaf
      if (tasked) {
	for (auto ch : child)
#pragma omp task firstprivate(ch) default(shared) untied final(depth >= params::task_recursion_cutoff_level-1) mergeable
	  ch->compress(F, strat, random_cols, Rrow, Rcol, ldr, Scol, Srow, lds, depth+1, false);
#pragma omp taskwait
	if (state!=COMPRESSED) for (auto ch : child) if (ch->state!=COMPRESSED) { return; }
      } else {
	for (auto ch : child) {
	  ch->compress(F, strat, random_cols, Rrow, Rcol, ldr, Scol, Srow, lds, depth, false);
	  if (state!=COMPRESSED) if (ch->state!=COMPRESSED) { return; }
	}
      }

      u_rows = v_rows = 0;
      for (auto ch : child) {
	integer_t* itmp = new integer_t[std::max(ch->v_rank,ch->u_rank)];
	scalar_t* stmp = new scalar_t[(random_cols-old_random_cols)*std::max(ch->u_rank,ch->v_rank)];
	if (ch->u_rank) { // get Rcol, Srow and Irow from child j
	  // move Rcol, be careful because this can overlap
	  if (Rcol+subset_begin+u_rows != Rcol+ch->subset_begin)
	    for (integer_t i=old_random_cols; i<random_cols; i++)
	      for (integer_t j=0; j<ch->u_rank; j++)
		Rcol[subset_begin+u_rows+j+i*ldr] = Rcol[ch->subset_begin+j+i*ldr];
	  for (integer_t i=old_random_cols; i<random_cols; i++)
	    for (integer_t j=0; j<ch->u_rank; j++)
	      stmp[j+(i-old_random_cols)*ch->u_rank] = Srow[ch->subset_begin+ch->locr[j]+i*lds];
	  blas::lacpy('A', ch->u_rank, random_cols_loc, stmp, ch->u_rank, Srow_loc+u_rows, lds);
	  for (integer_t i=0; i<ch->u_rank; i++) itmp[i] = indr[ch->subset_begin+ch->locr[i]];
	  for (integer_t i=0; i<ch->u_rank; i++) indr[u_rows+subset_begin+i] = itmp[i];
	  u_rows += ch->u_rank;
	}
	if (ch->v_rank) { // get Rrow, Scol and Icol from child j
	  // move Rrow, be careful because this can overlap
	  if (Rrow+subset_begin+v_rows != Rrow+ch->subset_begin)
	    for (integer_t i=old_random_cols; i<random_cols; i++)
	      for (integer_t j=0; j<ch->v_rank; j++)
		Rrow[subset_begin+v_rows+j+i*ldr] = Rrow[ch->subset_begin+j+i*ldr];
	  for (integer_t i=old_random_cols; i<random_cols; i++)
	    for (integer_t j=0; j<ch->v_rank; j++)
	      stmp[j+(i-old_random_cols)*ch->v_rank] = Scol[ch->subset_begin+ch->locc[j]+i*lds];
	  blas::lacpy('A', ch->v_rank, random_cols_loc, stmp, ch->v_rank, Scol_loc+v_rows, lds);
	  for (integer_t i=0; i<ch->v_rank; i++) itmp[i] = indc[ch->subset_begin+ch->locc[i]];
	  for (integer_t i=0; i<ch->v_rank; i++) indc[v_rows+subset_begin+i] = itmp[i];
	  v_rows += ch->v_rank;
	}
	delete[] itmp;
	delete[] stmp;
      }
      if (B==NULL) {
	integer_t Bsize = 0;
	for (auto ci : child) for (auto cj : child) if(ci != cj) Bsize += ci->u_rank*cj->v_rank;
	B = new scalar_t[Bsize];
      }
      integer_t Bpos = 0, jpos = 0;
      for (auto cj : child) { // loop over all combinations of children (nu_i, nu_j)
	integer_t ipos = 0;
	for (auto ci : child) {
	  if (ci != cj && ci->u_rank && cj->v_rank) {
	    if (state==UNTOUCHED) // extract A(I_{nu_i}, I_{nu_j})
	      F->extract(indr+subset_begin+ipos, ci->u_rank, indc+subset_begin+jpos, cj->v_rank, B+Bpos, ci->u_rank, depth);
	    if (!isroot) {
	      if (tasked) {
#pragma omp task default(shared) firstprivate(ci,cj,jpos,ipos,Bpos)
		gemm_omp_task('C', 'N', cj->v_rank, random_cols_loc, ci->u_rank, scalar_t(-1.0), // Scol = Scol - A(I_{nu_i}, I_{nu_j})* Rcol
			      B+Bpos, ci->u_rank, Rcol_loc+ipos, ldr, scalar_t(1.0), Scol_loc+jpos, lds, depth);
#pragma omp task default(shared) firstprivate(ci,cj,jpos,ipos,Bpos)
		gemm_omp_task('N', 'N', ci->u_rank, random_cols_loc, cj->v_rank, scalar_t(-1.0), // Srow = Srow - A(I_{nu_i}, I_{nu_j}) Rrow
			      B+Bpos, ci->u_rank, Rrow_loc+jpos, ldr, scalar_t(1.0), Srow_loc+ipos, lds, depth);
	      } else {
		blas::gemm('C', 'N', cj->v_rank, random_cols_loc, ci->u_rank, scalar_t(-1.0),
			   B+Bpos, ci->u_rank, Rcol_loc+ipos, ldr, scalar_t(1.0), Scol_loc+jpos, lds);
		blas::gemm('N', 'N', ci->u_rank, random_cols_loc, cj->v_rank, scalar_t(-1.0),
			   B+Bpos, ci->u_rank, Rrow_loc+jpos, ldr, scalar_t(1.0), Srow_loc+ipos, lds);
	      }
	    } Bpos += ci->u_rank * cj->v_rank;
	  } ipos += ci->u_rank;
	} jpos += cj->v_rank;
      }
      if (tasked) {
#pragma omp taskwait
      }
    }

    if (!isroot) {
      if (state!=COMPRESSED) {
	reallocate_workmemory(u_rows*random_cols, v_rows*random_cols);
	if (locr==NULL) {
	  locr = new int[2*(u_rows+v_rows)];
	  locc = locr + u_rows;
	  U_perm = locc + v_rows;
	  V_perm = U_perm + u_rows;
	}
	if (tasked) {
#pragma omp task default(shared)
	  ID<scalar_t,real_t,integer_t>(u_rows, random_cols, Srow+subset_begin, lds, U, work_u, U_perm, locr, u_rank, strat, depth);
	  ID<scalar_t,real_t,integer_t>(v_rows, random_cols, Scol+subset_begin, lds, V, work_v, V_perm, locc, v_rank, strat, depth);
#pragma omp taskwait
	} else {
	  ID<scalar_t,real_t>(u_rows, random_cols, Srow+subset_begin, lds, U, work_u, U_perm, locr, u_rank, strat, depth);
	  if (U==NULL) { state = PARTIALLY_COMPRESSED; return; }
	  ID<scalar_t,real_t>(v_rows, random_cols, Scol+subset_begin, lds, V, work_v, V_perm, locc, v_rank, strat, depth);
	}
	if (U==NULL || V==NULL) { state = PARTIALLY_COMPRESSED; return; }

	if (tasked) {
#pragma omp task default(shared)
	  {
	    apply_U(true, random_cols, Rcol+subset_begin, ldr, scalar_t(0.), work_u, u_rank, depth);
	    blas::lacpy('A', u_rank, random_cols, work_u, u_rank, Rcol+subset_begin, ldr);
	  }
	  apply_V(true, random_cols, Rrow+subset_begin, ldr, scalar_t(0.), work_v, v_rank, depth);
	  blas::lacpy('A', v_rank, random_cols, work_v, v_rank, Rrow+subset_begin, ldr);
#pragma omp taskwait
	} else {
	  apply_U(true, random_cols, Rcol+subset_begin, ldr, scalar_t(0.), work_u, u_rank, depth);
	  blas::lacpy('A', u_rank, random_cols, work_u, u_rank, Rcol+subset_begin, ldr);
	  apply_V(true, random_cols, Rrow+subset_begin, ldr, scalar_t(0.), work_v, v_rank, depth);
	  blas::lacpy('A', v_rank, random_cols, work_v, v_rank, Rrow+subset_begin, ldr);
	}
      } else {
	reallocate_workmemory(u_rank*random_cols, v_rank*random_cols);
	if (tasked) {
#pragma omp task default(shared)
	  {
	    apply_U(true, random_cols_loc, Rcol_loc, ldr, scalar_t(0.), work_u, u_rank, depth);
	    blas::lacpy('A', u_rank, random_cols_loc, work_u, u_rank, Rcol_loc, ldr);
	  }
	  apply_V(true, random_cols_loc, Rrow_loc, ldr, scalar_t(0.), work_v, v_rank, depth);
	  blas::lacpy('A', v_rank, random_cols_loc, work_v, v_rank, Rrow_loc, ldr);
#pragma omp taskwait
	} else {
	  apply_U(true, random_cols_loc, Rcol_loc, ldr, scalar_t(0.), work_u, u_rank, depth);
	  blas::lacpy('A', u_rank, random_cols_loc, work_u, u_rank, Rcol_loc, ldr);
	  apply_V(true, random_cols_loc, Rrow_loc, ldr, scalar_t(0.), work_v, v_rank, depth);
	  blas::lacpy('A', v_rank, random_cols_loc, work_v, v_rank, Rrow_loc, ldr);
	}
      }
    }
    state = COMPRESSED;
  }

} // end namespace strumpack

#endif //HSS_MATRIX_COMPRESS_HPP
