/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */
#ifndef HSS_MATRIX_EXTRACT_HPP
#define HSS_MATRIX_EXTRACT_HPP

#include <vector>
#include <numeric>

namespace strumpack {

  // Extract a single element from an HSS matrix. No reuse possible.
  template<typename scalar_t,typename real_t,typename integer_t> scalar_t
  HSSMatrix<scalar_t,real_t,integer_t>::get_ij(integer_t row, integer_t col) {
    if (child.empty()) return D[row + col * u_rows];

    integer_t a_dim = subset_end - subset_begin;
    scalar_t* e_col = new scalar_t[2*a_dim];
    scalar_t* e_row = &e_col[a_dim];

    for (integer_t i=0; i<a_dim; i++) e_col[i] = scalar_t(0.);
    for (integer_t i=0; i<a_dim; i++) e_row[i] = scalar_t(0.);

    e_col[col] = scalar_t(1.);
    matvec(e_col, a_dim, scalar_t(0.), e_row, a_dim, 1);
    scalar_t ij = e_row[row];

    delete[] e_col;
    return ij;
  }

  // the submatrix is ADDED to the values matrix
  // this is only for extraction from the update part (child[1]) of the matrix!!
  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::
  extract_submatrix(std::vector<integer_t>& rows, std::vector<integer_t>& cols,
		    std::vector<integer_t>& orig_rows, std::vector<integer_t>& orig_cols,
		    scalar_t* values, integer_t ld, integer_t dim_upd, integer_t dim_sep, int depth) {
    if (child[1]->child.empty()) {
      for (size_t c=0; c<cols.size(); c++)
	for (size_t r=0; r<rows.size(); r++)
	  values[orig_rows[r]+orig_cols[c]*ld] += child[1]->D[rows[r]-dim_sep+(cols[c]-dim_sep)*dim_upd];
      STRUMPACK_FLOPS((is_complex<scalar_t>()?2:1)*static_cast<long long int>(double(cols.size())*double(rows.size())));
      return;
    }
    depth *= 2; // this makes sure that not too many tasks are created for the extraction
    bool tasked = depth<params::task_recursion_cutoff_level;
    WorkmemTree<scalar_t,integer_t> w;
    w.allocate_children();
    if (tasked) {
      for (size_t i=0; i<child.size(); i++)
#pragma omp task firstprivate(i) default(shared) final(depth >= params::task_recursion_cutoff_level-1) mergeable
	child[1]->child[i]->extract_forward_loop(cols, w.c[i], depth+1);
#pragma omp taskwait
    } else {
      for (size_t i=0; i<child.size(); i++)
	child[1]->child[i]->extract_forward_loop(cols, w.c[i], depth);
    }
    for (size_t i=0; i<child.size(); i++) w.flag += w.c[i]->flag;
    child[1]->extract_backward_loop(rows, cols, orig_rows, orig_cols, values, ld, &w, true, depth);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::extract_forward_loop
  (std::vector<integer_t>& my_cols, WorkmemTree<scalar_t,integer_t>* w, int depth) {
    w->allocate_wmem(u_rank*my_cols.size(), v_rank*my_cols.size());
    w->flag = 0; // flag is used to count how many columns this node has
    bool tasked = depth<params::task_recursion_cutoff_level;

    if (child.empty()) { // leaf
      integer_t* Vind = new integer_t[v_rows];
      for (integer_t i=0; i<v_rows; i++) Vind[i] = i;
      for (integer_t i=v_rows-1; i>=0; i--) if (V_perm[i]!=i) std::swap(Vind[i], Vind[V_perm[i]]);
      for (size_t c=0; c<my_cols.size(); c++)
	if (my_cols[c] >= subset_begin && my_cols[c] < subset_end) {
	  blas::copy(v_rank, &V[Vind[my_cols[c]-subset_begin]], v_rows, &w->wmemv[w->flag*v_rank], 1);
	  w->flag++;
	}
      delete[] Vind;
    } else { // nonleaf
      w->allocate_children();
      if (tasked) {
	for (size_t i=0; i<child.size(); i++)
#pragma omp task firstprivate(i) default(shared) final(depth >= params::task_recursion_cutoff_level-1) mergeable
	  child[i]->extract_forward_loop(my_cols, w->c[i], depth+1);
#pragma omp taskwait
      } else {
	for (size_t i=0; i<child.size(); i++)
	  child[i]->extract_forward_loop(my_cols, w->c[i], depth);
      }
      for (size_t i=0; i<child.size(); i++) w->flag += w->c[i]->flag;
      if (w->flag == 0) return;

      if (v_rank) {
	scalar_t* tmp = new scalar_t[v_rows*w->flag];
	integer_t v_pos = 0;
	for (size_t i=0; i<child.size(); i++) {
	  integer_t col = 0;
	  integer_t ch_col = 0;
	  HSSMatrix<scalar_t,real_t,integer_t>* ch = child[i];
	  for (integer_t c=0; c<w->flag; c++) {
	    while (!(my_cols[col] >= subset_begin && my_cols[col] < subset_end)) col++;  // find the next column in parent
	    if (my_cols[col] >= ch->subset_begin && my_cols[col] < ch->subset_end) {     // if also in this child, then copy from y
	      std::copy(w->c[i]->wmemv+ch_col*ch->v_rank, w->c[i]->wmemv+ch_col*ch->v_rank+ch->v_rank, tmp+v_pos+c*v_rows);
	      ch_col++;
	    } else std::fill(tmp+v_pos+c*v_rows, tmp+v_pos+c*v_rows+ch->v_rank, scalar_t(0.));
	    col++;
	  }
	  v_pos += ch->v_rank;
	}
	apply_V(true, w->flag, tmp, v_rows, scalar_t(0.), w->wmemv, v_rank, depth);
	delete[] tmp;
      }
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  HSSMatrix<scalar_t,real_t,integer_t>::extract_backward_loop
  (std::vector<integer_t>& my_rows, std::vector<integer_t>& my_cols, std::vector<integer_t>& orig_rows,
   std::vector<integer_t>& orig_cols, scalar_t* values, integer_t ld,
   WorkmemTree<scalar_t,integer_t>* w, bool isroot, int depth) {
    bool tasked = depth<params::task_recursion_cutoff_level;

    if (child.empty()) { // leaf, cannot be root because that special case is already handled
      integer_t* Uind = new integer_t[u_rows];
      for (integer_t i=0; i<u_rows; i++) Uind[i] = i;
      for (integer_t i=u_rows-1; i>=0; i--) if (U_perm[i]!=i) std::swap(Uind[i], Uind[U_perm[i]]);

      // TODO make threadlocal??
      std::vector<integer_t> loc_rows; loc_rows.reserve(params::hss_minimum_partition_size);
      for (size_t r=0; r<my_rows.size(); r++)
	if (my_rows[r] >= subset_begin && my_rows[r] < subset_end)
	  loc_rows.push_back(r);

      if (!loc_rows.empty()) {
	scalar_t* U_i = new scalar_t[loc_rows.size()*u_rank+loc_rows.size()*my_cols.size()];
	scalar_t* loc_val = U_i + loc_rows.size()*u_rank;
	std::fill(U_i, U_i+loc_rows.size()*u_rank, scalar_t(0.));
	integer_t U_pos = 0;
	for (auto r : loc_rows) {
	  integer_t t = Uind[my_rows[r] - subset_begin];
	  if (t < u_rank) U_i[U_pos*u_rank+t] = scalar_t(1.); // in the diagonal part of U
	  else blas::copy(u_rank, U+t, u_rows, U_i+U_pos*u_rank, 1);
	  U_pos++;
	}
	if (u_rank)
	  blas::gemm('T', 'N', loc_rows.size(), my_cols.size(), u_rank, scalar_t(1.), U_i, u_rank,
		     w->wmemu, u_rank, scalar_t(0.), loc_val, loc_rows.size());
	else std::fill(loc_val, loc_val+loc_rows.size()*my_cols.size(), scalar_t(0.));

	for (size_t c=0; c<my_cols.size(); c++) {
	  U_pos = 0;
	  if (my_cols[c] >= subset_begin && my_cols[c] < subset_end) {
	    for (auto r : loc_rows)
	      values[orig_rows[r]+orig_cols[c]*ld] += D[my_rows[r]-subset_begin+(my_cols[c]-subset_begin)*u_rows]
		+ loc_val[(U_pos++)+c*loc_rows.size()];
	    STRUMPACK_FLOPS((is_complex<scalar_t>()?2:1)*static_cast<long long int>(2.*double(loc_rows.size())));
	  } else for (auto r : loc_rows)
		   values[orig_rows[r]+orig_cols[c]*ld] += loc_val[(U_pos++)+c*loc_rows.size()];
	}
	STRUMPACK_FLOPS((is_complex<scalar_t>()?2:1)*static_cast<long long int>(double(my_cols.size())*double(loc_rows.size())));
	delete[] U_i;
      }
      delete[] Uind;
    } else {
      scalar_t* tmp = new scalar_t[std::max(u_rows,v_rows)*my_cols.size()];
      if (!isroot && u_rank) {
	integer_t pos = 0;
	apply_U(false, my_cols.size(), w->wmemu, u_rank, scalar_t(0.), tmp, u_rows, depth);
	for (size_t i=0; i<child.size(); i++) {
	  blas::lacpy('A', child[i]->u_rank, my_cols.size(), tmp+pos, u_rows, w->c[i]->wmemu, child[i]->u_rank);
	  pos += child[i]->u_rank;
	}
      } else
	for (size_t i=0; i<child.size(); i++)
	  mat_zero(child[i]->u_rank, integer_t(my_cols.size()), w->c[i]->wmemu, child[i]->u_rank);
      integer_t Bpos = 0;
      for (size_t i=0; i<child.size(); i++)
	for (size_t j=0; j<child.size(); j++)
	  if (i != j) {
	    HSSMatrix<scalar_t,real_t,integer_t>* ci = child[i];
	    HSSMatrix<scalar_t,real_t,integer_t>* cj = child[j];
	    if (cj->u_rank && ci->v_rank) {
	      if (tasked) gemm_omp_task('N', 'N', cj->u_rank, w->c[i]->flag, ci->v_rank, scalar_t(1.0), &B[Bpos], cj->u_rank,
					w->c[i]->wmemv, ci->v_rank, scalar_t(0.), tmp, cj->u_rank, depth);
	      else blas::gemm('N', 'N', cj->u_rank, w->c[i]->flag, ci->v_rank, scalar_t(1.0), &B[Bpos], cj->u_rank,
			      w->c[i]->wmemv, ci->v_rank, scalar_t(0.), tmp, cj->u_rank);
	      Bpos += cj->u_rank * ci->v_rank;
	      integer_t col = 0;
	      for (size_t c=0; c<my_cols.size(); c++)
		if (my_cols[c] >= ci->subset_begin && my_cols[c] < ci->subset_end) {
		  blas::axpy(cj->u_rank, scalar_t(1.), &tmp[col*cj->u_rank], 1, &(w->c[j]->wmemu[c*cj->u_rank]), 1);
		  col++;
		}
	    }
	  }
      delete[] tmp;
      if (tasked) {
	for (size_t i=0; i<child.size(); i++)
#pragma omp task firstprivate(i) default(shared) final(depth >= params::task_recursion_cutoff_level-1) mergeable
	  child[i]->extract_backward_loop(my_rows, my_cols, orig_rows, orig_cols, values, ld, w->c[i], false, depth+1);
#pragma omp taskwait
      } else {
	for (size_t i=0; i<child.size(); i++)
	  child[i]->extract_backward_loop(my_rows, my_cols, orig_rows, orig_cols, values, ld, w->c[i], false, depth);
      }
    }
  }

} // end namespace strumpack

#endif
